/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2009 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#include "headers.h"
#include "forward.h"
#include "canvaswid.h"

#include "session.h"
#include "views.h"
#include "document.h"
#include "palette.h"

namespace Crow {

// Paint -----------------------------------------------------------------------

bool Paint::on_expose_event(GdkEventExpose * event) {
	PGdkWindow window= get_window();
	PGdkGC gc= Gdk::GC::create(window);
	Rectangle R= get_allocation();
	gc->set_rgb_fg_color(frame);
	window->draw_rectangle(gc, true, 0, 0, R.w, R.h);
	R.grow(1, 1, -2, -2);
	if(!R.empty()) {
		gc->set_rgb_fg_color(area);
		window->draw_rectangle(gc, true, 1, 1, R.w, R.h);
	}
	return true;
}

// WidgetCanvasEditorWidget ----------------------------------------------------

WidgetCanvasEditorWidget::WidgetCanvasEditorWidget() {

	setFrameNeeded(true);

	GAP= 5;
	scrollx= scrolly= 0;

	scroll.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);
	add(scroll);
	scroll.show();

	glass.set_above_child(true);
	glass.set_visible_window(false);
	glass.add_events(
		Gdk::POINTER_MOTION_MASK |
		Gdk::POINTER_MOTION_HINT_MASK |
		Gdk::BUTTON_PRESS_MASK |
		Gdk::BUTTON_RELEASE_MASK |
		Gdk::ENTER_NOTIFY_MASK |
		Gdk::LEAVE_NOTIFY_MASK
	);
	scroll.add(glass);
	dynamic_cast<Gtk::Viewport *>(scroll.get_child())->set_shadow_type(Gtk::SHADOW_NONE);
	glass.show();

	grabber.add_events(
		Gdk::POINTER_MOTION_MASK |
		Gdk::BUTTON_PRESS_MASK |
		Gdk::BUTTON_RELEASE_MASK |
		Gdk::ENTER_NOTIFY_MASK |
		Gdk::LEAVE_NOTIFY_MASK
	);
	glass.add(grabber);
	grabber.show();

	grabber.add(fixed);
	fixed.show();

	fixed.put(border, 0, 0);
	border.show();

	container.set_border_width(GAP);
	border.add(container);
	container.show();

	lastCursor= Gdk::LAST_CURSOR;

	scrollc= scroll.signal_size_allocate().connect(sigc::mem_fun(*this, &WidgetCanvasEditorWidget::setupState));

}

WidgetCanvasEditorWidget::~WidgetCanvasEditorWidget() {
	unsetWidget();
	clearPaintList();
}

void WidgetCanvasEditorWidget::setWidget(Gtk::Widget * widget) {
	if(widget != container.get_child()) {
		BinRemove(&container);
		container.add(*widget);
	}
}

Gtk::Widget * WidgetCanvasEditorWidget::getWidget() {
	return container.get_child();
}

void WidgetCanvasEditorWidget::unsetWidget() {
	BinRemove(&container);
}

void WidgetCanvasEditorWidget::setCursor(Gdk::CursorType cursor) {
	if(glass.get_window() && cursor!=lastCursor) {
		lastCursor= cursor;
		glass.get_window()->set_cursor(Gdk::Cursor(cursor));
	}
}

void WidgetCanvasEditorWidget::unsetCursor() {
	if(glass.get_window()) {
		lastCursor= Gdk::LAST_CURSOR;
		glass.get_window()->set_cursor();
	}
}

void WidgetCanvasEditorWidget::pack(const Point & size) {
	int nx= size.x <= 0 ? size.x : size.x + 2*getGAP();
	int ny= size.y <= 0 ? size.y : size.y + 2*getGAP();
	int ox, oy;
	container.get_size_request(ox, oy);
	if(ox!=nx || oy!=ny)
		container.set_size_request(nx, ny);
}

void WidgetCanvasEditorWidget::clearPaintList() {
	for(PaintList::iterator it= paintList.begin(); it!=paintList.end(); ++it)
		delete *it;
	paintList.clear();
}

void WidgetCanvasEditorWidget::beginPaint() {
	clearPaintList();
}

void WidgetCanvasEditorWidget::paint(const Rectangle & R, bool fill, const Gdk::Color & frame, const Gdk::Color & area) {
	if(fill) {
		Paint * paint= new Paint(frame, area);
		paint->set_size_request(R.w, R.h);
		fixed.put(*paint, R.x+GAP, R.y+GAP);
		paint->show();
		paintList.push_back(paint);
	} else {
		paint(Rectangle(R.x, R.y, R.w-1, 1), true, frame, area);
		paint(Rectangle(R.x+R.w-1, R.y, 1, R.h-1), true, frame, area);
		paint(Rectangle(R.x+1, R.y+R.h-1, R.w-1, 1), true, frame, area);
		paint(Rectangle(R.x, R.y+1, 1, R.h-1), true, frame, area);
	}
}

void WidgetCanvasEditorWidget::endPaint() {
}

void WidgetCanvasEditorWidget::loadState(PObject state) {
	PState S= PState::cast_dynamic(state);
	if(S) {
		scrollx= S->sx;
		scrolly= S->sy;
	}
}

void WidgetCanvasEditorWidget::setupState(Gtk::Allocation &) {
	scroll.get_hadjustment()->set_value(scrollx);
	scroll.get_vadjustment()->set_value(scrolly);
	scrollc.disconnect();
}

PObject WidgetCanvasEditorWidget::saveState() {
	PState S= NewRefPtr<State>();
	S->sx= scroll.get_hadjustment()->get_value();
	S->sy= scroll.get_vadjustment()->get_value();
	return S;
}

// WidgetCanvasEditor ----------------------------------------------------------

WidgetCanvasEditor::Location::Location() {
	clear();
}

void WidgetCanvasEditor::Location::clear() {
	widget= PNode();
	dx= dy= dw= dh= 0;
}

bool WidgetCanvasEditor::Location::manipulator() {
	return !empty() && !border() && !internal();
}

bool WidgetCanvasEditor::Location::border() {
	return !empty() && test(1, 1, 0, 0);
}

bool WidgetCanvasEditor::Location::internal() {
	return !empty() && test(0, 0, 0, 0);
}

bool WidgetCanvasEditor::Location::empty() {
	return !widget;
}

bool WidgetCanvasEditor::Location::test(int x, int y, int w, int h) {
	return dx==x && dy==y && dw==w && dh==h;
}

Gdk::CursorType WidgetCanvasEditor::Location::cursor() {
	CHECK(!empty());
	if(test(1, 1, -1, -1))
		return Gdk::TOP_LEFT_CORNER;
	if(test(0, 1, 0, -1))
		return Gdk::TOP_SIDE;
	if(test(0, 1, 1, -1))
		return Gdk::TOP_RIGHT_CORNER;
	if(test(1, 0, -1, 0))
		return Gdk::LEFT_SIDE;
	if(test(0, 0, 1, 0))
		return Gdk::RIGHT_SIDE;
	if(test(1, 0, -1, 1))
		return Gdk::BOTTOM_LEFT_CORNER;
	if(test(0, 0, 0, 1))
		return Gdk::BOTTOM_SIDE;
	if(test(0, 0, 1, 1))
		return Gdk::BOTTOM_RIGHT_CORNER;
	if(border())
		return Gdk::FLEUR;
	return Gdk::CIRCLE;
}

WidgetCanvasEditor::WidgetCanvasEditor() {
	GAP= 0;
	ML= 2;
	control= false;
	selectedOnPress= false;
	pasteNormal= true;
	button= -1;
	state= Selecting;
	interacting= false;
	stopInteracting= true;
	deleting= false;
}

void WidgetCanvasEditor::created() {
	PWidgetCanvasEditorWidget ew= getEditorWidget();
	GAP= ew->getGAP();
	ew->getGlass()->signal_enter_notify_event().connect(sigc::mem_fun(*this, &WidgetCanvasEditor::onMouseEnter));
	ew->getGlass()->signal_leave_notify_event().connect(sigc::mem_fun(*this, &WidgetCanvasEditor::onMouseLeave));
	ew->getGlass()->signal_motion_notify_event().connect(sigc::mem_fun(*this, &WidgetCanvasEditor::onMouseMove));
	ew->getGlass()->signal_button_press_event().connect(sigc::mem_fun(*this, &WidgetCanvasEditor::onMouseButtonPress));
	ew->getGlass()->signal_button_release_event().connect(sigc::mem_fun(*this, &WidgetCanvasEditor::onMouseButtonRelease));
	getManager()->connect_action_completed(sigc::mem_fun(*this, &WidgetCanvasEditor::onActionCompleted));
}

void WidgetCanvasEditor::load(bool reload) {
	PWidgetCanvasEditorWidget ew= getEditorWidget();
	Model * model= getManager()->getModel();
	PNode master= model->raiseMaster(getSession()->getNode1());
	if(!reload) {
		topMaster= master;
		ew->setWidget(getWidget(master));
		ew->pack(getDesignSize(master));
		getEditorWidget()->loadState(getManager()->getStates()->get(topMaster));
		beginSelecting();
	} else {
		CHECK(topMaster == master);
		ew->pack(getDesignSize(master));
	}
}

void WidgetCanvasEditor::deleted() {
	getManager()->getStates()->set(topMaster, getEditorWidget()->saveState());
	deleting= true;
	CHECK(state==Selecting || state==Pasting);
	if(state==Pasting)
		read(Cancel);
	CHECK(state==Selecting);
}

PNode WidgetCanvasEditor::getTopMaster() {
	return topMaster;
}

bool WidgetCanvasEditor::startPasting(const Glib::ustring & data, bool normal) {
	CHECK(state==Selecting || state==Pasting);
	pasteData= data;
	pasteNormal= normal;
	if(state==Selecting)
		read(Paste);
	return true;
}

void WidgetCanvasEditor::stopPasting() {
	CHECK(state==Pasting);
	read(Cancel);
}

void WidgetCanvasEditor::onActionCompleted() {
	showSelectionFrames();
}

void WidgetCanvasEditor::read(Command command) {

	switch(state) {

		case Selecting:
			switch(command) {

				case Move:
					selectingMove();
					break;

				case Press:
					testLocation(mouse);
					if(location.manipulator()) {
						state= Adjusting;
					} else if(!location.empty() && button==1) {
						state= Selecting1;
						selectingPress();
					}
					break;

				case Release:
					if(button==3)
						selectingMenu();
					break;

				case Paste:
					state= Pasting;
					beginPasting();
					break;

				default:;

			}
			break;

		case Selecting1:
			switch(command) {

				case Move:
					state= Dragging;
					beginDragging();
					draggingMove();
					break;

				case Press2:
					if(isInteractive(location.widget))
						interact();
					break;

				case Release:
					state= Selecting;
					selectingRelease();
					break;

				default:;

			}
			break;

		case Dragging:
			switch(command) {

				case Move:
					draggingMove();
					break;

				case Release:
					endDragging();
					state= Selecting;
					beginSelecting();
					break;

				default:;

			}
			break;

		case Pasting:
			switch(command) {

				case Move:
					pastingMove();
					break;

				case Leave:
					pastingLeave();
					break;

				case Press:
					state= Pasting1;
					pastingPress();
					break;

				case Cancel:
					endPasting();
					state= Selecting;
					pastingFinished();
					beginSelecting();
					break;

				default:;

			}
			break;

		case Pasting1:
			switch(command) {

				case Release:
					if(!control) {
						endPasting();
						state= Selecting;
						pastingFinished();
						beginSelecting();
					} else {
						state= Pasting;
						pastingMove();
					}
					break;

				default:;

			}
			break;

		case Adjusting:
			switch(command) {

				case Move:
					adjustingMove();
					break;

				case Release:
					endAdjusting();
					state= Selecting;
					beginSelecting();
					break;

				default:;

			}
			break;

		default:;

	}

}

void WidgetCanvasEditor::beginSelecting() {
	showSelectionFrames();
	getEditorWidget()->unsetCursor();
	//Glib::signal_idle().connect(sigc::bind_return(sigc::mem_fun(*this, &WidgetCanvasEditor::selectingMove), false));
}

void WidgetCanvasEditor::selectingMove() {
	testLocation(mouse);
	if(location.empty()) {
		getEditorWidget()->unsetCursor();
		updateStatus("");
	} else {
		if(location.manipulator() || location.border())
			getEditorWidget()->setCursor(location.cursor());
		else
			getEditorWidget()->unsetCursor();
		updateStatus(location.widget, getWidgetRect(location.widget).size());
	}
}

void WidgetCanvasEditor::selectingPress() {
	CHECK(!location.empty());
	selectedOnPress= getManager()->isSelected(location.widget);
	if(!selectedOnPress) {
		if(control) {
			Nodes selection= *getManager()->getSelection();
			selection.push_back(location.widget);
			getManager()->submitSelection(selection);
		} else {
			getManager()->submitSelection(Nodes(1, location.widget));
		}
	}
}

void WidgetCanvasEditor::selectingRelease() {
	if(selectedOnPress) {
		if(control) {
			Nodes selection= *getManager()->getSelection();
			CHECK(FindAndErase(selection, location.widget));
			getManager()->submitSelection(selection);
		} else {
			getManager()->submitSelection(Nodes(1, location.widget));
		}
	}
}

void WidgetCanvasEditor::selectingMenu() {
	testLocation(mouse);
	if(!location.empty()) {
		if(!getManager()->isSelected(location.widget))
			getManager()->submitSelection(Nodes(1, location.widget));
		popupMenu();
	}
}

void WidgetCanvasEditor::beginPasting() {
	getEditorWidget()->setCursor(Gdk::CROSSHAIR);
	frames.clear();
	paintFrames();
}

void WidgetCanvasEditor::pastingMove() {
	FreePlace fplace;
	PNode widget= findWidget(mouse);
	bool found= false;
	bool changed;
	if(widget) {
		Point wp= toLocal(widget, mouse);
		found= getWidgetView(widget)->findFreePlace(wp, &fplace);
	}
	if(found)
		changed= createSingleFrame(Rectangle(
			toGlass(widget, fplace.area.point()),
			fplace.area.size()
		));
	else {
		changed= !frames.empty();
		frames.clear();
	}
	if(changed)
		paintFrames("wce_vc");
}

void WidgetCanvasEditor::pastingLeave() {
	if(!frames.empty()) {
		frames.clear();
		paintFrames();
	}
}

void WidgetCanvasEditor::pastingPress() {
	PNode widget= findWidget(mouse);
	if(widget) {
		Point wp= toLocal(widget, mouse);
		Model * model= getManager()->getModel();

		getManager()->begin(pasteNormal ? umNormal : umPaste);

		Document document;
		Nodes loadedNodes;
		bool loaded= document.load(model, pasteData, loadedNodes);
		bool widgets= true;
		if(loaded) {

			Nodes masters= model->selectMasters(loadedNodes);
			for(Nodes::iterator it=masters.begin(); it!=masters.end(); ++it)
				if(
					!Palette::get()->isWidget((*it)->getType()) ||
					GetEntry<EntityEntry>((*it)->getType())->hasFlag(efTopMaster)
				) {
					widgets= false;
					break;
				}

			if(widgets) {
				FreePlaces fplaces= getWidgetView(widget)->findFreePlaces(wp, masters, wp);
				if(!fplaces.empty()) {

					int i=0;
					for(Nodes::iterator it= masters.begin(); it!=masters.end(); ++it, ++i)
						getWidgetView(widget)->createChild(fplaces[i].place, *it);

					getManager()->commit(masters);
					return;
				}
			}

		}

		getManager()->rollback();

		read(Release);
		button= -1;
		std::string msg;
		if(loaded)
			msg= widgets ? "Cannot find free place(s)" : "Non-widget objects could be pasted to root only";
		else
			msg= "Pasting failed: " + *document.getError();
		getController()->getContext()->showMessage(msg);

	}

}

void WidgetCanvasEditor::endPasting() {
	pasteNormal= true;
	pasteData.clear();
}

void WidgetCanvasEditor::beginDragging() {
	getEditorWidget()->setCursor(Gdk::PLUS);
	frames.clear();
}

void WidgetCanvasEditor::draggingMove() {
	const Nodes & reduced= *getManager()->getReducedSelection();
	FreePlaces fplaces;
	PNode widget= findWidget(mouse, true);
	Frames F;
	if(widget) {
		Point wp= toLocal(widget, mouse);
		Point wp0= toLocal(widget, press);
		fplaces= getWidgetView(widget)->findFreePlaces(wp, reduced, wp0);
	}
	if(!fplaces.empty()) {
		int i= 0;
		for(Nodes::const_iterator it=reduced.begin(); it!=reduced.end(); ++it, ++i)
			F[*it]= Rectangle(
				toGlass(widget, fplaces[i].area.point()),
				fplaces[i].area.size()
			);
	} else {
		if(widget)
			F[widget]= getWidgetRect(widget);
		else
			for(Nodes::const_iterator it=reduced.begin(); it!=reduced.end(); ++it)
				F[*it]= getWidgetRect(*it);
	}
	if(F != frames) {
		frames= F;
		paintFrames(fplaces.empty() ? "wce_wc" : "wce_vc");
	}
}

void WidgetCanvasEditor::endDragging() {
	frames.clear();
	const Nodes & reduced= *getManager()->getReducedSelection();
	FreePlaces fplaces;
	PNode widget= findWidget(mouse, true);
	if(widget) {
		Point wp= toLocal(widget, mouse);
		Point wp0= toLocal(widget, press);
		fplaces= getWidgetView(widget)->findFreePlaces(wp, reduced, wp0);
	}
	if(!fplaces.empty()) {
		getManager()->begin();
		int i;

		std::vector<Nodes> props;
		props.resize(fplaces.size());
		i= 0;
		for(Nodes::const_iterator it=reduced.begin(); it!=reduced.end(); ++it, ++i)
			if(widget->getType() == getManager()->getModel()->findParent(*it)->getType()) {
				PNode child= getManager()->getModel()->findChildNode(*it);
				Nodes out= *child->getOut();
				for(Nodes::iterator it=out.begin(); it!=out.end(); ++it) {
					Property * prop= getManager()->getController()->findViewProperty(*it);
					if(!prop->hasFlags(sfChildDef)) {
						getManager()->getModel()->cut(*it);
						props[i].push_back(*it);
					}
				}
			}

		i= 0;
		for(Nodes::const_iterator it=reduced.begin(); it!=reduced.end(); ++it, ++i) {
			PNode child= getWidgetView(widget)->createChild(fplaces[i].place, *it);
			for(Nodes::iterator it=props[i].begin(); it!=props[i].end(); ++it)
				getManager()->getModel()->paste(*it, child);
		}

		getManager()->commit();
	}
}

void WidgetCanvasEditor::adjustingMove() {

	Point k= mouse - press;

	Rectangle rect= getWidgetRect(location.widget);
	rect.grow(
		k.x*location.dx,
		k.y*location.dy,
		k.x*location.dw,
		k.y*location.dh,
		1, 1
	);

	frames[location.widget]= rect;

	paintFrames();

	updateStatus(location.widget, rect.size());
}

void WidgetCanvasEditor::endAdjusting() {
	if(frames[location.widget] != getWidgetRect(location.widget)) {
		if(location.widget==getSession()->getNode1())
			setDesignSize(frames[location.widget].size());
		else if(isXYChild(location.widget))
			setXYChildPlacement(location.widget, frames[location.widget]);
	}
}

void WidgetCanvasEditor::showSelectionFrames() {
	if((state==Selecting || state==Selecting1) && !deleting)
		Glib::signal_idle().connect(sigc::bind_return(sigc::mem_fun(*this, &WidgetCanvasEditor::showSelectionFramesIdle), false));
}

void WidgetCanvasEditor::showSelectionFramesIdle() {
	bool changed= createSelectionFrames();
	if(changed)
		paintFrames();
}

bool WidgetCanvasEditor::createSelectionFrames() {
	Frames F;
	const Nodes & selection= *getManager()->getSelection();
	Model * model= getManager()->getModel();
	for(Nodes::const_iterator it= selection.begin(); it!=selection.end(); ++it) {
		PNode node= *it;
		if(isActiveWidget(node) && model->raiseParent(*it) == getSession()->getNode1())
			F[node]= getWidgetRect(node);
	}
	if(F != frames) {
		frames= F;
		return true;
	}
	return false;
}

bool WidgetCanvasEditor::createSingleFrame(const Rectangle & rect) {
	Frames F;
	F[PNode()]= rect;
	if(F != frames) {
		frames= F;
		return true;
	}
	return false;
}

void WidgetCanvasEditor::paintFrames(const std::string & frameColorName) {

	Gdk::Color frameColor;
	if(frameColorName.empty())
		frameColor= GetBlack(getEditorWidget()->get_style());
	else
		frameColor= getController()->getContext()->getKeyValue<Gdk::Color>(frameColorName);
	Gdk::Color manipColor= GetWhite(getEditorWidget()->get_style());

	getEditorWidget()->beginPaint();

	for(Frames::iterator it= frames.begin(); it!=frames.end(); ++it) {
		const Rectangle & R= (*it).second;

		Rectangle frame= R;
		frame.grow(-1, -1, 2, 2);
		getEditorWidget()->paint(frame, false, frameColor, Gdk::Color());

		if(state==Selecting || state==Selecting1 || state==Adjusting)
			for(int X=0; X<3; ++X)
				for(int Y=0; Y<3; ++Y)
					if(X!=1 || Y!=1) {
						Rectangle manip= getManipulatorRect(R, X, Y);
						getEditorWidget()->paint(manip, true, frameColor, manipColor);
					}

	}

	getEditorWidget()->endPaint();

}

void WidgetCanvasEditor::testLocation(const Point & p) {
	CHECK(state==Selecting || state==Selecting1);

	location.clear();

	for(Frames::iterator it= frames.begin(); it!=frames.end(); ++it)
		for(int X=0; X<3; ++X)
			for(int Y=0; Y<3; ++Y)
				if(X!=1 || Y!=1)
					if(p.inside(getManipulatorRect((*it).second, X, Y))) {
						location.widget= (*it).first;
						location.dx= (2-X)/2;
						location.dy= (2-Y)/2;
						location.dw= X-1;
						location.dh= Y-1;
						if(location.widget==getSession()->getNode1() && !getController()->getModel()->findMaster(location.widget)) {
							if(X==0 || Y==0) {
								location.dx= 1;
								location.dy= 1;
								location.dw= 0;
								location.dh= 0;
							}
						} else if(!isXYChild(location.widget)) {
							location.dx= 0;
							location.dy= 0;
							location.dw= 0;
							location.dh= 0;
						}
						return;
					}

	for(Frames::iterator it= frames.begin(); it!=frames.end(); ++it) {
		Rectangle inner= (*it).second;
		Rectangle outer= (*it).second;
		inner.grow(2, 2, -4, -4);
		outer.grow(-1, -1, 2, 2);
		if(p.inside(outer) && !p.inside(inner)) {
			location.widget= (*it).first;
			location.dx= 1;
			location.dy= 1;
			return;
		}
	}

	location.widget= findWidget(p);

}

PNode WidgetCanvasEditor::findWidget(const Point & p, bool ignoreSelection) {

	if(!isActiveWidget(getSession()->getNode1()))
		return PNode();

	PNode widget;
	PNode child= getSession()->getNode1();

	if(p.inside(getWidgetRect(child))) {

		do {

			if(ignoreSelection && getManager()->isSelected(child))
				break;

			widget= child;
			child= findChild(widget, p);

		} while(child);

	}

	return widget;
}

PNode WidgetCanvasEditor::findChild(PNode widget, const Point & p) {
	Model * model= getManager()->getModel();
	PGtkWidgetView view= getWidgetView(widget);
	if(toLocal(widget, p).inside(view->getChildArea())) {
		Nodes children= model->getChildren(widget, false);
		for(Nodes::const_iterator it= children.begin(); it!=children.end(); ++it) {
			PChild child= getController()->getView<ChildView>(model->findChildNode(*it))->getObject();
			if(view->isTopChild(child) && p.inside(getWidgetRect(*it)))
				return *it;
		}
	}
	return PNode();
}

PNode WidgetCanvasEditor::findCompositeChild(PNode widget, const Point & p) {
	Model * model= getManager()->getModel();
	PGtkWidgetView view= getWidgetView(widget);
	if(p.inside(getWidgetRect(widget))) {
		Nodes nested= model->getNested(widget, true);
		Nodes children= model->getChildren(widget, false);
		Nodes slaves= model->getSlaves(widget, false);
		slaves.splice(slaves.end(), nested, nested.begin(), nested.end());
		children.sort();
		slaves.sort();
		Nodes nodes;
		std::set_difference(slaves.begin(), slaves.end(), children.begin(), children.end(), std::back_inserter(nodes));
		for(Nodes::const_iterator it= nodes.begin(); it!=nodes.end(); ++it)
			if(
				Palette::get()->isWidget((*it)->getType()) &&
				view->isActiveWidget(getWidget(*it)) &&
				p.inside(getWidgetRect(*it))
			)
				return *it;
	}
	return PNode();
}

bool WidgetCanvasEditor::isActiveWidget(PNode node) {

	PNode current= node;

	for(;;) {
		PNode master= getManager()->getModel()->findMaster(current);
		if(!master)
			return true;

		PEntityView view= getController()->getView(master);

		if(!view->isActiveWidget(getWidget(current)))
			return false;

		current= master;
	}

}

bool WidgetCanvasEditor::isInteractive(PNode node) {
	return getWidgetView(node)->isInteractive();
}

Gtk::Widget * WidgetCanvasEditor::getWidget(PNode node) {
	Any value= getController()->getView(node)->getValue();
	Gtk::Widget * widget= dynamic_cast<Gtk::Widget *>(GetPtr(value->getObject()));
	return widget;
}

Rectangle WidgetCanvasEditor::getWidgetRect(PNode node) {
	int width= getWidget(node)->get_allocation().get_width();
	int height= getWidget(node)->get_allocation().get_height();
	Point p= toGlass(node, Point(0, 0));
	return Rectangle(p.x, p.y, width, height);
}

Rectangle WidgetCanvasEditor::getManipulatorRect(const Rectangle & R, int X, int Y) {
	Rectangle rect;
	rect.x= R.x-1-ML + X*(R.w+1)/2;
	rect.y= R.y-1-ML + Y*(R.h+1)/2;
	rect.w= 2*ML+1;
	rect.h= 2*ML+1;
	return rect;
}

Point WidgetCanvasEditor::toGlass(PNode node, const Point & p) {
	Point r= Transform(p, getWidget(node), getEditorWidget()->getGlass());
	return r - Point(GAP, GAP);
}

Point WidgetCanvasEditor::toLocal(PNode node, const Point & p) {
	Point r= Transform(p, getEditorWidget()->getGlass(), getWidget(node));
	return r + Point(GAP, GAP);
}

PGtkWidgetView WidgetCanvasEditor::getWidgetView(PNode widget) {
	PEntityView view= getController()->getView(widget);
	return PGtkWidgetView::cast_dynamic(view);
}

bool WidgetCanvasEditor::isXYChild(PNode widget) {
	return getManager()->isXYChild(widget);
}

Point WidgetCanvasEditor::getDesignSize(PNode node) {
	Model * model= getManager()->getModel();
	PNode ds= model->find(node, "design-size");
	Point size= ds ? *ds->getValue()->getPoint() : Point(-1, -1);
	return size;
}

void WidgetCanvasEditor::setDesignSize(const Point & size) {
	Model * model= getManager()->getModel();
	getManager()->begin();
	model->setScalar(getSession()->getNode1(), "design-size", CAny::createPoint(size));
	getManager()->commit();
}

void WidgetCanvasEditor::setXYChildPlacement(PNode widget, const Rectangle & r) {
	Model * model= getManager()->getModel();
	PNode child= model->findChildNode(widget);
	PNode parent= model->findParent(widget);
	Point shift= dynamic_cast<XYContainer *>(PGtkContainerView::cast_dynamic(getWidgetView(parent))->getContainer())->getShift();
	getManager()->begin();
	model->setScalar(widget, "size-request", CAny::createPoint(r.size()));
	model->setScalar(child, "xy", CAny::createPoint(-shift + toLocal(parent, r.point())));
	getManager()->commit();
}

bool WidgetCanvasEditor::onMouseEnter(GdkEventCrossing * event) {
	if(event->window==getEditorWidget()->getGlass()->get_window()->gobj() && !interacting) {
		read(Enter);
		mouseMove(event->x - GAP, event->y - GAP);
	}
	return true;
}

bool WidgetCanvasEditor::onMouseLeave(GdkEventCrossing * event) {
	if(event->window==getEditorWidget()->getGlass()->get_window()->gobj() && !interacting) {
		updateStatus("");
		read(Leave);
	}
	return true;
}

bool WidgetCanvasEditor::onMouseMove(GdkEventMotion * event) {
	mouseMove(event->x - GAP, event->y - GAP);
	int x, y;
	Gdk::ModifierType state;
	getEditorWidget()->getGlass()->get_window()->get_pointer(x, y, state);
	return true;
}

bool WidgetCanvasEditor::onMouseButtonPress(GdkEventButton * event) {
	if(event->type==GDK_BUTTON_PRESS)
		buttonPress(event->x - GAP, event->y - GAP, (Gdk::ModifierType)event->state, event->button);
	else if(event->type==GDK_2BUTTON_PRESS)
		buttonPress2(event->button);
	return true;
}

bool WidgetCanvasEditor::onMouseButtonRelease(GdkEventButton * event) {
	buttonRelease(event->x - GAP, event->y - GAP, (Gdk::ModifierType)event->state, event->button);
	return true;
}

void WidgetCanvasEditor::mouseMove(int x, int y) {
	if(!interacting) {
		mouse= Point(x, y);
		read(Move);
	}
}

void WidgetCanvasEditor::buttonPress(int x, int y, Gdk::ModifierType state, int mb) {
	if(button==-1 && !interacting) {
		button= mb;
		press= mouse= Point(x, y);
		control= state & Gdk::CONTROL_MASK;
		read(Press);
	}
}

void WidgetCanvasEditor::buttonPress2(int mb) {
	if(button==1 && mb==1 && !interacting)
		read(Press2);
}

void WidgetCanvasEditor::buttonRelease(int x, int y, Gdk::ModifierType state, int mb) {
	if(button==mb && button!=-1 && !interacting) {
		mouse= Point(x, y);
		control= state & Gdk::CONTROL_MASK;
		read(Release);
		button= -1;
	}
}

void WidgetCanvasEditor::customDoEvent(GdkEvent * event, gpointer data) {
	static_cast<WidgetCanvasEditor *>(data)->doEvent(event);
}

void WidgetCanvasEditor::doEvent(GdkEvent * event) {
	CHECK(!stopInteracting);
	stopInteracting= event && event->type==GDK_BUTTON_RELEASE && event->button.button==1;
	if(stopInteracting)
		gdk_event_handler_set((GdkEventFunc)gtk_main_do_event, NULL, NULL);
	gtk_main_do_event(event);
}

gboolean WidgetCanvasEditor::interactionButtonRelease(
	GtkWidget * widget,
	GdkEventButton * event,
	bool * stopInteracting
) {
	if(event->button==1)
		*stopInteracting= true;
	return FALSE;
}

void WidgetCanvasEditor::interact() {

	CHECK(!interacting && !location.empty());

	if(findCompositeChild(location.widget, mouse))
		return;

	interacting= true;
	stopInteracting= false;

	if(frames!=Frames()) {
		frames.clear();
		paintFrames();
	}

	getEditorWidget()->getGlass()->set_above_child(false);
	GdkWindow * window= Gdk::Display::get_default()->get_window_at_pointer()->gobj();

	GtkWidget * widget= NULL;
	gdk_window_get_user_data(window, (void **)(void *)&widget);
	CHECK(GTK_IS_WIDGET(widget));
	gulong id= g_signal_connect(widget, "button-release-event", G_CALLBACK(&WidgetCanvasEditor::interactionButtonRelease), &stopInteracting);

	gdk_pointer_grab(window, FALSE, GdkEventMask(GDK_POINTER_MOTION_MASK | GDK_BUTTON_RELEASE_MASK), NULL, NULL, GDK_CURRENT_TIME);

	synthesize(window, GDK_ENTER_NOTIFY);
	synthesize(window, GDK_BUTTON_PRESS);

	//gdk_event_handler_set(&WidgetCanvasEditor::customDoEvent, this, NULL);

	while(!stopInteracting)
		gtk_main_iteration();

	g_signal_handler_disconnect(widget, id);

	synthesize(window, GDK_LEAVE_NOTIFY);

	//while(Gtk::Main::events_pending())
	//	Gtk::Main::iteration();

	//gdk_flush();

	getEditorWidget()->getGlass()->set_above_child(true);
	gdk_pointer_ungrab(GDK_CURRENT_TIME);

	getManager()->begin();
	getManager()->getModel()->tagModified(location.widget);
	getManager()->commit();

	synthesize(getEditorWidget()->getGrabber()->get_window()->gobj(), GDK_BUTTON_RELEASE);
	interacting= false;
}

void WidgetCanvasEditor::synthesize(GdkWindow * window, GdkEventType type) {

	CHECK(interacting);

	int rx, ry;
	gdk_window_get_origin(window, &rx, &ry);
	int px, py;
	GdkModifierType mask;
	gdk_window_get_pointer(window, &px, &py, &mask);

	GdkEvent * event= gdk_event_new(type);
	g_object_ref(window);
	event->any.window= window;
	event->any.send_event= true;

	switch(type) {

		case GDK_BUTTON_PRESS:
		case GDK_BUTTON_RELEASE:
			event->button.time= GDK_CURRENT_TIME;
			event->button.x= px;
			event->button.y= py;
			event->button.axes= NULL;
			event->button.state= GDK_BUTTON1_MASK;
			event->button.button= 1;
			event->button.device= Gdk::Display::get_default()->get_core_pointer()->gobj();
			event->button.x_root= rx + px;
			event->button.y_root= ry + py;
			break;

		case GDK_ENTER_NOTIFY:
		case GDK_LEAVE_NOTIFY:
			g_object_ref(window);
			event->crossing.subwindow= window;
			event->crossing.time= GDK_CURRENT_TIME;
			event->crossing.x= px;
			event->crossing.y= py;
			event->crossing.x_root= rx + px;
			event->crossing.y_root= ry + py;
			event->crossing.mode= GDK_CROSSING_NORMAL;
			event->crossing.detail= GDK_NOTIFY_VIRTUAL;
			event->crossing.focus= false;
			event->crossing.state= GDK_BUTTON1_MASK;
			break;

		default:
			CHECK(false);

	}

	Gdk::Display::get_default()->put_event(event);
	gdk_event_free(event);

}

void WidgetCanvasEditor::updateStatus(const Glib::ustring & text) {
	statusChanged(text);
}

void WidgetCanvasEditor::updateStatus(PNode widget, const Point & size) {
	const std::string & str= *Palette::get()->getEntry(widget->getType())->getTypeName();
	updateStatus(
		*widget->getName() + ", " + str +
		", width: " + ToString(size.x) +
		" height: " + ToString(size.y));
}

}
