/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2009 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#include "headers.h"
#include "forward.h"
#include "controller.h"

#include "document.h"
#include "views.h"
#include "palette.h"
#include "widgets.h"

namespace Crow {

// Context ---------------------------------------------------------------------

bool Context::hasKey(const std::string & key) {
	return getSettings()->find(key) != getSettings()->end();
}

void Context::ensureKey(const std::string & key, const std::string & val) {
	getSettings()->insert(std::make_pair(key, val));
}

// Controller ------------------------------------------------------------------

Controller::Controller() {
	context= NULL;
}

Controller::~Controller() {
	clear();
	CHECK(views.empty());
}

void Controller::clearState() {
	model.clearState();
	touched.clear();
}

void Controller::begin(UpdateMode um) {
	model.begin(um);
}

void Controller::rollback() {
	model.rollback();
}

void Controller::commit() {
	synchronize();
	model.commit();
}

void Controller::undo() {
	model.begin(umSeek);
	model.seek(model.getPointer()-1);
	synchronize();
	model.commit();
}

void Controller::redo() {
	model.begin(umSeek);
	model.seek(model.getPointer()+1);
	synchronize();
	model.commit();
}

bool Controller::canUndo() {
	return model.getPointer() > 0;
}

bool Controller::canRedo() {
	return model.getPointer() < model.getMarkCount()-1;
}

LoadResult Controller::load(const Glib::ustring & data) {
	Document document;
	Nodes loaded;
	clear();
	clearState();
	model.begin(umLoad);
	model.tagModified();
	bool success= document.load(&model, data, loaded);
	LoadResult result;
	if(success) {
		commit();
		result= model.testChange(ctAppend) ? lrSuccess : lrChanged;
	} else {
		rollback();
		result= lrFailed;
	}
	if(result!=lrFailed)
		return result;
	else
		throw RuntimeError(*document.getError());
}

void Controller::clear() {
	model.clear();
	synchronize();
}

Glib::ustring Controller::save(const Nodes & nodes) {
	Document document;
	return document.save(
		&model, nodes,
		sigc::mem_fun(this, &Controller::willSave)
	);
}

Property * Controller::findViewProperty(PNode node) {

	PEntityView view;
	const std::string * pname;

	PNode owner= node->getOwner();

	CHECK(owner && owner!=model.getRoot());

	if(owner->getRole()==nrEntity) {
		view= getView(owner);
		pname= node->getName();
	} else {
		CHECK(owner->getRole()==nrVector);
		view= getView(owner->getOwner());
		pname= owner->getName();
	}

	Property * property= view->findProperty(*pname);
	CHECK(property);

	return property;
}

NodeTest Controller::willSave(PNode node) {
	PNode owner= node->getOwner();
	if(owner==model.getRoot())
		return NodeTest(true, NULL);
	Property * property= findViewProperty(node);
	if(owner->getRole()==nrVector)
		return NodeTest(true, property);
	if(property->isReadOnly() || property->hasFlags(sfNoSave))
		return NodeTest(false, property);
	if(node->getRole()==nrVector)
		return NodeTest(!node->getOut()->empty(), property);
	if(node->getRole()==nrScalar) {
		Any dvalue= property->getDefault();
		bool save= !dvalue || !dvalue->equals(node->getValue()) || !node->getMeta()->empty();
		return NodeTest(save, property);
	}
	if(node->getRole()==nrLink)
		return NodeTest(node->getLink(), property);
	CHECK(node->getRole()==nrEntity);
	Any dvalue= property->getDefault();
	CHECK(dvalue);
	CHECK(isVoidEntity(dvalue));
	return NodeTest(true, property);
}

bool Controller::canDefault(PNode node) {
	PNode owner= node->getOwner();
	if(owner->getRole()==nrVector)
		return node->getRole()==nrLink && node->getLink();
	CHECK(owner!=model.getRoot());
	Property * property= findViewProperty(node);
	if(!property->canEdit())
		return false;
	if(node->getRole()==nrVector)
		return !node->getOut()->empty();
	if(node->getRole()==nrScalar) {
		Any dvalue= property->getDefault();
		return dvalue && (!dvalue->equals(node->getValue()) || !node->getMeta()->empty());
	}
	if(node->getRole()==nrLink)
		return node->getLink();
	CHECK(node->getRole()==nrEntity);
	Any dvalue= property->getDefault();
	CHECK(dvalue);
	CHECK(isVoidEntity(dvalue));
	return true;
}

PEntityView Controller::findView(Any value) {
	PEntityView result;
	for(ViewMap::iterator it= views.begin(); it!=views.end(); ++it) {
		PEntityView view= it->second;
		if(value->equals(view->getValue())) {
			CHECK(!result);
			result= view;
		}
	}
	return result;
}

PEntityView Controller::getView(PNode node) {
	ViewMap::iterator it= views.find(node);
	return it==views.end() ? PEntityView() : it->second;
}

void Controller::touch(PNode node) {
	CHECK(model.getUpdateMode() != umNone);
	touched.insert(node);
}

bool Controller::isTouched(PNode node) {
	return touched.find(node) != touched.end();
}

void Controller::objectDestroyed(PGlibObject object) {
	destroyed.push_back(object);
}

Controller::Emitters Controller::getEmitters() {
	Emitters result;
	for(ViewMap::const_iterator it= views.begin(); it!=views.end(); ++it) {
		GlibObjectView * view= dynamic_cast<GlibObjectView *>(GetPtr(it->second));
		if(view) {
			Property * psignals= view->findProperty("signals");
			if(psignals) {
				Any value= psignals->getInert();
				if(value) {
					const AnyVector & vec= *value->getVector();
					for(int i=0; i<SIGNED(vec.size()); ++i) {
						const Emitter & emitter= *vec[i]->get<Emitter>();
						std::string signature= view->findSignal(emitter.signame);
						EmitterInfo emt;
						emt.view= view;
						emt.signature= signature;
						emt.emitter= emitter;
						result.insert(std::make_pair(emitter.mapname, emt));
					}
				}
			}
		}
	}
	return result;
}

bool Controller::modelGetScalar(PNode node, Any * result, PNode owner, const std::string & name, StateFlags sf) {
	CHECK(node->getState() == sf);
	if(node->isNoUndo() && model.getUpdateMode()==umSeek)
		return false;
	*result= node->getValue();
	return true;
}

bool Controller::modelGetVector(PNode node, Any * result, PNode owner, const std::string & name, TypeId base, StateFlags sf) {

	CHECK(node->getRole() == nrVector);

	AnyVector vec;
	const Nodes & nodes= *node->getOut();
	vec.resize(nodes.size());

	for(Nodes::const_iterator it= nodes.begin(); it!=nodes.end(); ++it) {
		const std::string & skey= *(*it)->getName();
		int key= FromString<int>(skey);
		CHECK(0 <= key && key < SIGNED(vec.size()) && !vec[key]);

		Any val;
		PNode elem= model.find(node, skey);
		if(isEntity(base))
			CHECK(modelGetEntity(elem, &val, node, skey, base, NULL));
		else
			CHECK(modelGetScalar(elem, &val, node, skey, sf));

		if(elem->getRole()==nrLink || elem->getRole()==nrScalar)
			CHECK(elem->getState() == sf);
		else
			CHECK(elem->getRole()==nrEntity);

		vec[key]= val;
	}

	*result= CAny::createVector(vec);

	return true;
}

bool Controller::modelGetEntity(PNode node, Any * result, PNode owner, const std::string & name, TypeId base, Property * property) {

	if(property) {
		if(node->getRole()==nrLink)
			CHECK(node->getState() == property->getNodeState());
		else
			CHECK(node->getRole()==nrEntity);
	} else
		CHECK(owner->getRole()==nrVector);

	if(node->getRole() == nrLink) {
		PNode target= node->getLink();

		if(target) {
			CHECK(Palette::get()->typeIsA(target->getType(), base));
			*result= getView(target)->getValue();
		} else
			*result= CAny::createObject();

	} else {
		CHECK(node->getRole() == nrEntity);
		CHECK(Palette::get()->typeIsA(node->getType(), base));

		PEntityView view;

		if(property && property->isReadOnly()) {

			view= getView(node);
			Any existing= property->getValue();

			if(view) {

				if(!view->getValue()->equals(existing))
					return false;

			} else {

				if(isVoidEntity(existing))
					return false;

				view= ensureView(node, existing);

			}

		} else {
			view= ensureView(node);
		}

		if(property)
			property->configureView(view);
		view->update();

		*result= view->getValue();

	}

	return true;
}

PNode Controller::modelSetScalar(PNode owner, const std::string & name, Any value, Property * property) {
	PNode node= model.find(owner, name);
	if(node) {
		CHECK(node->getState() == property->getNodeState());
		model.setScalar(node, value);
		if(property->isReadOnly())
			model.setMeta(node, "");
	} else
		node= model.createScalar(owner, name, value, property->getNodeState());
	return node;
}

PNode Controller::modelSetVector(PNode owner, TypeId type, const std::string & name, Any value) {
	PNode node= model.find(owner, name);
	if(!node)
		node= model.createVector(owner, name);
	CHECK(node->getRole() == nrVector);
	if(value->getVector()->empty())
		model.vectorClear(node);
	else
		CHECK(value->getVector()->size() == node->getOut()->size());
	if(isEntity(type)) {
		const Nodes & nodes= *node->getOut();
		for(Nodes::const_iterator it= nodes.begin(); it!=nodes.end(); ++it)
			if((*it)->getRole()!=nrLink)
				getView(*it)->write();
	}
	return node;
}

PNode Controller::modelSetEntity(PNode owner, const std::string & name, Any value, Property * property) {

	PNode node= model.find(owner, name);

	if(node) {

		if(node->getRole()==nrLink)
			CHECK(node->getState() == property->getNodeState());
		else
			CHECK(node->getRole()==nrEntity);

		if(!isVoidEntity(value)) {

			if(node->getRole() == nrLink) {
				PNode target= node->getLink();

				if(!target) {
					node= makeNodeForValue(node, owner, name, value, property);
				} else {
					if(!value->equals(getView(target)->getValue()))
						node= makeNodeForValue(node, owner, name, value, property);
				}

			} else {
				CHECK(node->getRole() == nrEntity);
				PEntityView view= getView(node);
				CHECK(view);

				if(value->equals(view->getValue())) {
					property->configureView(view);
					view->write();
				} else {
					model.remove(node);
					node= makeNodeForValue(PNode(), owner, name, value, property);
				}

			}

		} else {

			if(node->getRole() == nrLink) {
				PNode target= node->getLink();
				if(target) {
					if(!node->isWeakLink())
						model.remove(target);
					else
						model.clearLink(node);
				}
			} else {
				CHECK(node->getRole() == nrEntity);
				model.remove(node);
				node= model.createLink(owner, name, property->getNodeState());
			}

		}

	} else {

		if(!isVoidEntity(value))
			node= makeNodeForValue(PNode(), owner, name, value, property);
		else
			node= model.createLink(owner, name, property->getNodeState());

	}

	return node;
}

PNode Controller::makeNodeForValue(PNode old, PNode owner, const std::string & name, Any value, Property * property) {
	PNode node;
	PEntityView foundv= findView(value);
	PNode found= foundv ? foundv->getModelNode() : PNode();
	if(found) {
		node= makeLink(old, owner, name, property);
		model.setLink(node, found);
	} else {
		if(property->hasFlags(sfLinkOnly)) {
			node= makeLink(old, owner, name, property);
		} else {
			if(old)
				model.remove(old);
			node= model.createEntity(owner, value->getType(), name);
			ensureAndWrite(node, value, property);
		}
	}
	return node;
}

PNode Controller::makeLink(PNode old, PNode owner, const std::string & name, Property * property) {
	if(old) {
		CHECK(old->getRole() == nrLink);
		CHECK(old->getState() == property->getNodeState());
		model.clearLink(old);
		return old;
	} else
		return model.createLink(owner, name, property->getNodeState());
}

void Controller::ensureAndWrite(PNode node, Any value, Property * property) {
	PEntityView view= ensureView(node, value);
	property->configureView(view);
	view->write();
}

PEntityView Controller::ensureView(PNode node, Any val) {
	CHECK(!val || node->getType() == val->getType());
	PEntityView view= getView(node);
	CHECK(!(view && val) || view->getValue()->equals(val));
	if(!view) {
		EntityEntry * entry= GetEntry<EntityEntry>(node->getType());
		view= entry->createView();
		setView(node, view);
		view->init(node, this, val);
	}
	return view;
}

void Controller::synchronize() {
	bool cycle= model.getUpdateMode()==umNormal;
	int pass= 0;
	int passPrint= 2;
	bool retry= false;

	do {
		++pass;
		CHECK(pass <= 10);
		if(pass == passPrint)
			std::cout << "Synchronization passes: " << passPrint << std::flush;
		else if(pass > passPrint)
			std::cout << ", " << pass << std::flush;

		Nodes updated= update();
		model.clearLinksCleared();
		write(updated);

		retry= model.getLinksCleared();
		CHECK(cycle || !retry);

	} while(retry);

	if(pass >= passPrint)
		std::cout << "." << std::endl;

	sweepViews();
}

Nodes Controller::getOrderedModified() {
	const Nodes & out= *model.getRoot()->getOut();
	Nodes result;
	for(Nodes::const_iterator it= out.begin(); it!=out.end(); ++it)
		if((*it)->isModified())
			result.push_back(*it);

	struct tmp { static bool less(const PNode & a, const PNode & b) {
		int pr_a= Palette::get()->getEntry(a->getType())->getPriority();
		int pr_b= Palette::get()->getEntry(b->getType())->getPriority();
		return pr_a < pr_b;
	} };

	result.sort(&tmp::less);

	return result;
}

Nodes Controller::update() {
	model.setReadOnly(true);

	Nodes modified= getOrderedModified();

	context->sweep();

	int updated_count= 0;

	Nodes::iterator i1= modified.begin();
	Nodes::iterator i2= modified.begin();
	while(i2!=modified.end()) {
		Nodes::iterator in= i2;
		++in;
		int p2= GetEntry((*i2)->getType())->getPriority();
		int pn= in==modified.end() ? 100 : GetEntry((*in)->getType())->getPriority();
		if(pn > p2) {
			for(Nodes::iterator ii= i1; ; ++ii) {
				++updated_count;
				CHECK(!(*ii)->isInactive());
				ensureView(*ii);
				if(ii==i2)
					break;
			}
			for(Nodes::iterator ii= i1; ; ++ii) {
				CHECK(!(*ii)->isInactive());
				getView(*ii)->update();
				if(ii==i2)
					break;
			}
			i1= in;
		}
		i2= in;
	}

	CHECK(updated_count==SIGNED(modified.size()));

	model.setReadOnly(false);
	return modified;
}

void Controller::write(const Nodes & updated) {
	for(Nodes::const_iterator it= updated.begin(); it!=updated.end(); ++it)
		if(!(*it)->isInactive())
			getView(*it)->write();
}

void Controller::setView(PNode node, PEntityView view) {
	CHECK(view);
	CHECK(!getView(node));
	views[node]= view;
}

void Controller::removeView(PNode node) {
	ViewMap::iterator it= views.find(node);
	CHECK(it!=views.end());
	it->second->done();
	views.erase(it);
}

void * Controller::destroyNotify(void * data) {
	Palette::get()->incrementLeaks(-1);
	return NULL;
}

void Controller::sweepViews() {
	for(ViewMap::iterator it= views.begin(); it!=views.end(); ) {
		if(it->first->isInactive()) {
			it->second->done();
			it= EraseAndNext(views, it);
		} else
			++it;
	}
	while(!destroyed.empty()) {
		Glib::Object * object= GetPtr(destroyed.back());
		CHECK(!dynamic_cast<Gtk::Window *>(object));
		object->add_destroy_notify_callback(NULL, &Controller::destroyNotify);
		destroyed.pop_back();
		Palette::get()->incrementLeaks(1);
	}
}

bool Controller::isVoidEntity(Any value) {
	return value->equals(CAny::createObject());
}

bool Controller::isEntity(TypeId type) {
	return Palette::get()->isEntity(type);
}

bool Controller::isWidget(PNode node) {
	return Palette::get()->isWidget(node->getType());
}

}
