/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2009 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#include "headers.h"
#include "forward.h"
#include "crow_impl.h"

#include <guiloader.h>

#include "views2.h"
#include "palette.h"
#include "designer.h"

#include "../service/marshal.h"

extern "C" {

CROW_API int crow_version(void) {
	return CROW_VERSION;
}

CROW_API void crow_enter_check(const char * msg) {
	guiloader_enter_check(msg);
	Crow::CrowInit::enter(msg ? msg : "");
}

CROW_API void crow_leave(void) {
	Crow::CrowInit::leave();
	guiloader_leave();
}

CROW_API GQuark crow_error_quark(void) {
	static GQuark quark= 0;
	if(!quark)
		quark= g_quark_from_static_string("crow_error");
	return quark;
}

CROW_API GType crow_error_get_type(void) {
	static GType etype= 0;
	if(etype == 0) {
		static const GEnumValue values[]= {
			{ CROW_ERROR_RUNTIME, "CROW_ERROR_RUNTIME", "Runtime" },
			{ 0, NULL, NULL }
		};
		etype= g_enum_register_static("CrowError", values);
	}
	return etype;
}

CROW_API GType crow_designer_get_type(void) {
	static GType object_type= 0;
	if(object_type==0) {
		static const GTypeInfo object_info= {
			sizeof(CrowDesignerClass),
			NULL,
			NULL,
			(GClassInitFunc)crow_designer_class_init,
			NULL,
			NULL,
			sizeof(CrowDesigner),
			0,
			(GInstanceInitFunc)crow_designer_init,
		};
		object_type= g_type_register_static(G_TYPE_OBJECT, "CrowDesigner", &object_info, GTypeFlags(0));
	}
	return object_type;
}

void crow_designer_class_init(CrowDesignerClass * meta) {
	GObjectClass * object_class;
	meta->parent_class= static_cast<GObjectClass *>(g_type_class_peek_parent(meta));
	object_class= G_OBJECT_CLASS(meta);
	object_class->finalize= crow_designer_finalize;
	g_type_class_add_private(object_class, sizeof(CrowDesignerPrivate));

	g_signal_new(
		"actions_updated",
		G_OBJECT_CLASS_TYPE(object_class),
		G_SIGNAL_RUN_LAST, 0,
		NULL, NULL,
		crow_marshal_VOID__INT64,
		G_TYPE_NONE, 1,
		G_TYPE_INT64
	);
	g_signal_new(
		"status_updated",
		G_OBJECT_CLASS_TYPE(object_class),
		G_SIGNAL_RUN_LAST, 0,
		NULL, NULL,
		g_cclosure_marshal_VOID__STRING,
		G_TYPE_NONE, 1,
		G_TYPE_STRING
	);
	g_signal_new(
		"info_updated",
		G_OBJECT_CLASS_TYPE(object_class),
		G_SIGNAL_RUN_LAST, 0,
		NULL, NULL,
		g_cclosure_marshal_VOID__STRING,
		G_TYPE_NONE, 1,
		G_TYPE_STRING
	);
	g_signal_new(
		"show_url",
		G_OBJECT_CLASS_TYPE(object_class),
		G_SIGNAL_RUN_LAST, 0,
		NULL, NULL,
		g_cclosure_marshal_VOID__STRING,
		G_TYPE_NONE, 1,
		G_TYPE_STRING
	);
	g_signal_new(
		"modal_enter",
		G_OBJECT_CLASS_TYPE(object_class),
		G_SIGNAL_RUN_LAST, 0,
		NULL, NULL,
		g_cclosure_marshal_VOID__VOID,
		G_TYPE_NONE, 0
	);
	g_signal_new(
		"modal_leave",
		G_OBJECT_CLASS_TYPE(object_class),
		G_SIGNAL_RUN_LAST, 0,
		NULL, NULL,
		g_cclosure_marshal_VOID__VOID,
		G_TYPE_NONE, 0
	);

}

void crow_designer_init(CrowDesigner * obj) {
	CrowDesignerPrivate * priv= CROW_DESIGNER_GET_PRIVATE(obj);
	priv->designer= new Crow::DesignerImpl;
	priv->designer->connect_actions_updated(sigc::bind<CrowDesigner *>(sigc::ptr_fun(crow_designer_emit_actions_updated), obj));
	priv->designer->connect_status_updated(sigc::bind<CrowDesigner *>(sigc::ptr_fun(crow_designer_emit_status_updated), obj));
	priv->designer->connect_info_updated(sigc::bind<CrowDesigner *>(sigc::ptr_fun(crow_designer_emit_info_updated), obj));
	priv->designer->connect_show_url(sigc::bind<CrowDesigner *>(sigc::ptr_fun(crow_designer_emit_show_url), obj));
	priv->designer->connect_modal_enter(sigc::bind<CrowDesigner *>(sigc::ptr_fun(crow_designer_emit_modal_enter), obj));
	priv->designer->connect_modal_leave(sigc::bind<CrowDesigner *>(sigc::ptr_fun(crow_designer_emit_modal_leave), obj));
}

void crow_designer_finalize(GObject * object) {
  CrowDesigner * designer= CROW_DESIGNER(object);
  CrowDesignerClass * meta= CROW_DESIGNER_GET_CLASS(designer);
	CrowDesignerPrivate * priv= CROW_DESIGNER_GET_PRIVATE(designer);
	delete priv->designer;
	(*meta->parent_class->finalize)(object);
}

void crow_designer_emit_actions_updated(CrowActions changed, CrowDesigner * obj) {
	g_signal_emit_by_name(obj, "actions-updated", changed);
}

void crow_designer_emit_status_updated(const Glib::ustring & str, CrowDesigner * obj) {
	g_signal_emit_by_name(obj, "status-updated", str.c_str());
}

void crow_designer_emit_info_updated(const Glib::ustring & str, CrowDesigner * obj) {
	g_signal_emit_by_name(obj, "info-updated", str.c_str());
}

void crow_designer_emit_show_url(const Glib::ustring & str, CrowDesigner * obj) {
	g_signal_emit_by_name(obj, "show-url", str.c_str());
}

void crow_designer_emit_modal_enter(CrowDesigner * obj) {
	g_signal_emit_by_name(obj, "modal-enter");
}

void crow_designer_emit_modal_leave(CrowDesigner * obj) {
	g_signal_emit_by_name(obj, "modal-leave");
}

CROW_API CrowDesigner * crow_designer_new(void) {
	return static_cast<CrowDesigner *>(g_object_new(CROW_TYPE_DESIGNER, NULL));
}

CROW_API void crow_designer_apply_settings(CrowDesigner * obj, GHashTable * settings) {
	CrowDesignerPrivate * priv= CROW_DESIGNER_GET_PRIVATE(obj);
	priv->designer->apply_settings(Crow::HashTableToSettings(settings));
}

CROW_API void crow_designer_store_settings(CrowDesigner * obj, GHashTable * settings) {
	CrowDesignerPrivate * priv= CROW_DESIGNER_GET_PRIVATE(obj);
	Crow::Settings map= Crow::HashTableToSettings(settings);
	priv->designer->store_settings(&map);
	g_hash_table_foreach_remove(settings, reinterpret_cast<GHRFunc>(&gtk_true), NULL);
	Crow::SettingsToHashTable(map, settings);
}

CROW_API void crow_designer_clear(CrowDesigner * obj) {
	CrowDesignerPrivate * priv= CROW_DESIGNER_GET_PRIVATE(obj);
	priv->designer->clear();
}

CROW_API gboolean crow_designer_load(CrowDesigner * obj, const char * data, GError ** error) {
	try {
		CrowDesignerPrivate * priv= CROW_DESIGNER_GET_PRIVATE(obj);
		priv->designer->load(data);
		return TRUE;
	} catch(const Glib::Exception & x) {
		g_set_error(error, CROW_ERROR, CROW_ERROR_RUNTIME, x.what().c_str());
		return FALSE;
	}
}

CROW_API char * crow_designer_save(CrowDesigner * obj) {
	CrowDesignerPrivate * priv= CROW_DESIGNER_GET_PRIVATE(obj);
	Glib::ustring data= priv->designer->save();
	return g_strdup(data.c_str());
}

CROW_API char * crow_designer_save_header(CrowDesigner * obj) {
	CrowDesignerPrivate * priv= CROW_DESIGNER_GET_PRIVATE(obj);
	Glib::ustring data= priv->designer->save_header();
	return g_strdup(data.c_str());
}

CROW_API void crow_designer_saved(CrowDesigner * obj) {
	CrowDesignerPrivate * priv= CROW_DESIGNER_GET_PRIVATE(obj);
	priv->designer->saved();
}

CROW_API GtkWidget * crow_designer_get_widget(CrowDesigner * obj) {
	CrowDesignerPrivate * priv= CROW_DESIGNER_GET_PRIVATE(obj);
	return priv->designer->get_widget()->gobj();
}

CROW_API CrowActions crow_designer_get_enabled_actions(CrowDesigner * obj) {
	CrowDesignerPrivate * priv= CROW_DESIGNER_GET_PRIVATE(obj);
	return priv->designer->get_enabled_actions();
}

CROW_API void crow_designer_run_action(CrowDesigner * obj, CrowActions action) {
	CrowDesignerPrivate * priv= CROW_DESIGNER_GET_PRIVATE(obj);
	priv->designer->run_action(action);
}

}

namespace Crow {

// CrowInit --------------------------------------------------------------------

CrowInit * CrowInit::get() {
	CHECK(ptr());
	return ptr();
}

void CrowInit::enter(const std::string & msg) {
	if(!ptr()) {
		ptr()= new CrowInit();
		ptr()->init();
	} else
		++ptr()->refcounter;
	ptr()->snapshot(msg);
}

void CrowInit::leave() {
	CHECK(ptr() && ptr()->refcounter > 0);
	ptr()->check();
	--ptr()->refcounter;
	if(ptr()->refcounter == 0) {
		delete ptr();
		ptr()= NULL;
	}
}

CrowInit * & CrowInit::ptr() {
	static CrowInit * crowInit= NULL;
	return crowInit;
}

CrowInit::CrowInit() {
	CHECK(GTK_MAJOR_VERSION*1000+GTK_MINOR_VERSION <= gtk_major_version*1000+gtk_minor_version);
	CHECK(GUILOADER_VERSION/1000 <= guiloader_version()/1000);
	refcounter= 1;
	palette= new Palette();
}

CrowInit::~CrowInit() {
	delete palette;
}

void CrowInit::init() {
	palette->init();
}

void CrowInit::snapshot(const std::string & msg) {
	labels.push_back(msg);
	if(!msg.empty())
		snapshots.push_back(objects);
}

void CrowInit::check() {

	std::string msg= labels.back();
	labels.pop_back();

	if(!msg.empty()) {

		#ifdef CROW_SANDBOX

			const Set & snapshot= snapshots.back();

			CHECK(std::includes(objects.begin(), objects.end(), snapshot.begin(), snapshot.end()));

			std::list<Crow::Object *> leaked;
			std::set_difference(
				objects.begin(), objects.end(),
				snapshot.begin(), snapshot.end(),
				std::back_inserter(leaked)
			);

			int leaks= leaked.size();

			if(leaks > 0) {

				std::cerr << "[Possible leaks in " << msg << ": " << ToString(leaks) << "]" << std::endl;

				while(!leaked.empty()) {
					Crow::Object * obj= leaked.front();
					std::string cname= GetTypeName(typeid(*obj));
					std::cerr << "---------" << cname << ", " << obj->refcounter;

					CAny * any= dynamic_cast<CAny *>(obj);
					if(any && any->getType()) {
						std::cerr << ", "
							<< *palette->getEntry(any->getType())->getTypeName()
							<< "=" << any->toString();
					}

					Node * node= dynamic_cast<Node *>(obj);
					if(node) {
						std::cerr
							<< ", name=" << *node->getName()
							<< ", out=" << ((node->getRole()==nrScalar || node->getRole()==nrLink) ? 0 : node->getOut()->size())
							<< ", in=" << node->getIn()->size();
					}

					std::cerr << std::endl;
					leaked.pop_front();
				}

			}

		#else

			int snapshot= snapshots.back();
			int leaks= objects - snapshot;
			CHECK(leaks >= 0);

		#endif

		if(leaks != 0) {
			Glib::ustring str= "Possible leaks in " + msg + ": " + ToString(leaks);
			#ifndef CROW_SANDBOX
				std::cerr << str << std::endl;
			#endif
			//ShowMessage(str);
		}

		snapshots.pop_back();

	}

}

void CrowInit::objectCreated(Crow::Object * object) {
	#ifdef CROW_SANDBOX
		CHECK(objects.insert(object).second);
	#else
		++objects;
	#endif
}

void CrowInit::objectDeleted(Crow::Object * object) {
	#ifdef CROW_SANDBOX
		CHECK(objects.erase(object)==1);
	#else
		--objects;
	#endif
}

}
