/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2009 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#include "headers.h"
#include "forward.h"
#include "document.h"

#include "model.h"
#include "palette.h"
#include "crow_impl.h"

namespace Crow {

Document::Document() {
	model= NULL;
}

Glib::ustring Document::save(Model * mdl, const Nodes & nodes, const NodeSave & nsave) {

	model= mdl;
	nodeSet= model->growSlaves(nodes);
	nodeSave= nsave;

	std::ostringstream doc;
	doc << "<?xml version='1.0' encoding='UTF-8'?>\n";
	doc << "<gui namespace='GTK+' version='" << ToString(CROW_VERSION_GUIXML) << "'>\n";

	NodesSorted sorted(nodeSet.begin(), nodeSet.end());
	CHECK(sorted.size()==nodeSet.size());
	for(NodesSorted::iterator it=sorted.begin(); it!=sorted.end(); ++it)
		doc << save(*it, 1);

	doc << "</gui>\n";

	nodeSet.clear();
	nodeSave= NodeSave();
	model= NULL;

	return doc.str();
}

Glib::ustring Document::save(PNode node, int indent) {
	std::ostringstream child;
	Glib::ustring tabs(indent, '\t');

	NodeTest test= nodeSave(node);

	Glib::ustring meta= *node->getMeta();

	std::ostringstream attr;
	if(node->getOwner()->getRole()!=nrVector)
		attr << " name='" << *node->getName() << "'";
	if(node->getRole()!=nrLink && node->getRole()!=nrVector)
		attr << " type='" << *Palette::get()->getEntry(node->getType())->getTypeName() << "'";
	if(!meta.empty())
		attr << " meta='" << XmlEncode(meta, true) << "'";

	switch(node->getRole()) {

		case nrScalar: {
			if(test.first) {
				child << tabs << "<scalar" << attr.str() << ">";

				if(test.second->hasFlags(sfXml)) {

					Glib::ustring str= node->getValue()->toString();
					child << "<![CDATA[\n";

					StringList lines= Split(str, "\n");
					for(StringList::iterator it= lines.begin(); it!=lines.end(); ++it)
						if(!it->empty())
							child << tabs << '\t' << *it << '\n';

					child << tabs << "]]>";

				} else
					child << XmlEncode(node->getValue()->toString());

				child << "</scalar>\n";
			}
			break;
		}

		case nrVector: {
			if(test.first) {
				child << tabs << "<vector" << attr.str() << ">\n";
				NodesSorted sorted(node->getOut()->begin(), node->getOut()->end());
				CHECK(sorted.size()==node->getOut()->size());
				for(NodesSorted::const_iterator it= sorted.begin(); it!=sorted.end(); ++it)
					child << save(*it, indent+1);
				child << tabs << "</vector>\n";
			}
			break;
		}

		case nrEntity: {
			child << tabs << "<entity" << attr.str() << ">\n";
			NodesSorted sorted(node->getOut()->begin(), node->getOut()->end());
			CHECK(sorted.size()==node->getOut()->size());
			std::ostringstream domain;
			for(NodesSorted::const_iterator it= sorted.begin(); it!=sorted.end(); ++it)
				domain << save(*it, indent+1);
			if(!test.first && domain.str().empty())
				child.str("");
			else
				child << domain.str() << tabs << "</entity>\n";
			break;
		}

		case nrLink: {
			if(test.first) {
				PNode target= node->getLink();
				child << tabs << "<link" << attr.str();
				if(target && nodeSet.find(target)!=nodeSet.end())
					child << ">" << *target->getName() << "</link>\n";
				else {
					CHECK(!target || node->isWeakLink());
					child << "/>\n";
				}
			}
			break;
		}

		default:
			CHECK(false);

	}

	return child.str();
}

bool Document::load(Model * mdl, const Glib::ustring & data, Nodes & loaded) {

	model= mdl;
	error= "";
	bool success= true;

	try {

		try {
			Parser::parse(this, data);
		} catch(const Glib::Error & x) {
			if(x.what()=="upgrade") {
				Glib::ustring upgraded= GuiUpgrade().upgrade(data);
				std::cerr << upgraded << std::flush;
				Parser::parse(this, upgraded);
			} else
				throw;
		}

		CHECK(stack.empty());

		for(Links::iterator it= links.begin(); it!=links.end(); ++it) {
			std::string name= names[it->second];
			PNode target= model->findObject(name);
			if(target)
				model->setLink(it->first, target);
		}

		loaded= objects;

	} catch(const Glib::Error & x) {
		error= x.what();
		success= false;
	}

	stack.clear();
	names.clear();
	links.clear();
	objects.clear();

	model= NULL;

	return success;
}

StateFlags Document::getNodeState(const std::string & name) {
	Stack::iterator it= stack.end();
	--it;
	std::string property;
	if(it->node->getRole()==nrEntity)
		property= name;
	else {
		CHECK(it->node->getRole()==nrVector);
		property= *it->node->getName();
		--it;
		CHECK(it->node->getRole()==nrEntity);
	}
	return Palette::get()->getEntry<EntityEntry>(it->node->getType())->getNodeState(property);
}

void Document::onStartElement(const Glib::ustring & element, const Glib::Markup::Parser::AttributeMap & attributes) {

	if(element == "gui") {

		Glib::Markup::Parser::AttributeMap::const_iterator it;
		it= attributes.find("namespace");
		std::string nspace= it==attributes.end() ? "GTK+" : it->second;
		if(nspace!="GTK+")
			RaiseError("GuiXml namespace check failed");
		it= attributes.find("version");
		std::string version= it==attributes.end() ? "" : it->second;
		checkVersion(version);

		stack.push_back(SEntry(model->getRoot(), ""));

	} else {

		++stack.back().leafs;

		std::string name;
		std::string type;
		Glib::ustring meta;

		PNode node;

		for(Glib::Markup::Parser::AttributeMap::const_iterator it= attributes.begin(); it!=attributes.end(); ++it)
			if(it->first == "name")
				name= it->second;
			else if(it->first == "type")
				type= it->second;
			else if(it->first == "meta")
				meta= it->second;
			else
				RaiseError("unknown attribute '" + it->first + "'");

		if(stack.size()==1) {
			std::string unique= model->makeUnique(name);
			names[name]= unique;
			name= unique;
		} else if(stack.back().node->getRole()==nrVector) {
			CHECK(name.empty());
			name= ToString(stack.back().leafs - 1);
		}

		if(element=="scalar")
			node= model->createScalar(stack.back().node, FindType(type), name, getNodeState(name));
		else if(element=="vector")
			node= model->createVector(stack.back().node, name);
		else if(element=="entity")
			node= model->createEntity(stack.back().node, FindType(type), name);
		else if(element=="link")
			node= model->createLink(stack.back().node, name, getNodeState(name));
		else
			RaiseError("unknown element '" + element + "'");

		model->setMeta(node, meta);

		if(stack.size()==1)
			objects.push_back(node);

		stack.push_back(SEntry(node, ""));

	}

}

void Document::onEndElement(const Glib::ustring & name) {
	PNode node= stack.back().node;
	const Glib::ustring & text= stack.back().text;
	if(node->getRole()==nrScalar) {
		Any value= CAny::createFromString(node->getType(), text);
		model->setScalar(node, value);
	} else if(node->getRole()==nrLink) {
		links.push_back(std::make_pair(node, text));
	}
	stack.pop_back();
}

void Document::onText(const Glib::ustring & text) {
	stack.back().text += text;
}

void Document::onPassthrough(const Glib::ustring & text) {
	std::string str= text;
	if(Begins(str, std::string("<![CDATA[")) && Ends(str, std::string("]]>"))) {
		std::string sub= str.substr(9, str.size()-9-3);
		StringList lines= Split(sub, "\r\n");
		int skip= INT_MAX;
		for(StringList::iterator it= lines.begin(); it!=lines.end(); )
			if(!Trim(*it).empty()) {
				skip= std::min(skip, checked_cast<int>(it->find_first_not_of('\t')));
				++it;
			} else
				it= lines.erase(it);
		std::ostringstream shifted;
		for(StringList::iterator it= lines.begin(); it!=lines.end(); ++it) {
			int size= it->size();
			CHECK(skip < size);
			shifted << it->substr(skip, size - skip) << '\n';
		}
		onText(shifted.str());
	}
}

void Document::checkVersion(const std::string & vstr) {
	int version= FromString<int>(vstr);
	if(CROW_VERSION_GUIXML == version)
		return;
	if(CROW_VERSION_GUIXML < version)
		RaiseError("gui version check failed");
	if(version <= GuiNeedsUpgrade)
		throw Upgrade();
}

bool Document::idorder::operator () (PNode n1, PNode n2) {
	return LessSplit(*n1->getName(), *n2->getName());
}

// GuiUpgrade ------------------------------------------------------------------

Glib::ustring GuiUpgrade::upgrade(const Glib::ustring & data) {

	try {
		Parser::parse(this, data);
	} catch(const Glib::Error & x) {
		std::cerr << "UI parsing failed: " << x.what() << std::endl;
		return data;
	}

	switch(version) {

		case 1:
			upgrade1();

		case 2:
			upgrade2();

		case 3:
		case 4:
			upgrade4();

		case 5:
			upgrade5();

		case 6:
			upgrade6();

		case 7:
		case 8:
			upgrade8();

		case 9:
			upgrade9();

	}

	return write();
}

void GuiUpgrade::onStartElement(const Glib::ustring & element, const Glib::Markup::Parser::AttributeMap & attributes) {

	if(element == "gui") {

		Glib::Markup::Parser::AttributeMap::const_iterator it= attributes.find("version");
		std::string vstr= it==attributes.end() ? "" : it->second;
		version= FromString<int>(vstr);
		model.push_back(Node());
		stack.push_back(&model.front());

	} else {

		Node node;

		for(Glib::Markup::Parser::AttributeMap::const_iterator it= attributes.begin(); it!=attributes.end(); ++it)
			if(it->first == "name")
				node.name= it->second;
			else if(it->first == "type")
				node.type= it->second;
			else if(it->first == "flags") {
				StringList flags= Split(it->second, "|");
				std::copy(flags.begin(), flags.end(), std::back_inserter(node.flags));
			} else if(it->first == "meta")
				node.meta= it->second;
			else
				RaiseError("unknown attribute '" + it->first + "'");

		node.element= element;
		node.owner= stack.back();
		model.push_back(node);
		stack.back()->domain.push_back(&model.back());
		stack.push_back(&model.back());

	}

}

void GuiUpgrade::onEndElement(const Glib::ustring & name) {
	stack.pop_back();
}

void GuiUpgrade::onText(const Glib::ustring & text) {
	stack.back()->text += text;
}

void GuiUpgrade::onPassthrough(const Glib::ustring & text) {
	std::string str= text;
	if(Begins(str, std::string("<![CDATA[")) && Ends(str, std::string("]]>"))) {
		Glib::ustring sub= str.substr(9, str.size()-9-3);
		stack.back()->cdata += sub;
	}
}

Glib::ustring GuiUpgrade::write() {

	std::ostringstream doc;
	doc << "<?xml version='1.0' encoding='UTF-8'?>\n";
	doc << "<gui namespace='GTK+' version='" << ToString(CROW_VERSION_GUIXML) << "'>\n";

	std::list<Node *> & nodes= model.front().domain;
	for(std::list<Node *>::iterator it=nodes.begin(); it!=nodes.end(); ++it)
		doc << write(*it, 1);

	doc << "</gui>\n";

	return doc.str();
}

Glib::ustring GuiUpgrade::write(Node * node, int indent) {

	std::ostringstream child;
	Glib::ustring tabs(indent, '\t');

	std::ostringstream attr;
	attr << " name='" << node->name << "'";
	if(!node->type.empty())
		attr << " type='" << node->type << "'";
	if(!node->meta.empty())
		attr << " meta='" << XmlEncode(node->meta, true) << "'";

	if(node->element=="scalar" || node->element=="link") {
		child << tabs << "<" << node->element << attr.str() << ">";
		if(!node->cdata.empty())
			child << "<![CDATA[" << node->cdata << "]]>";
		else
			child << XmlEncode(node->text);
		child << "</" << node->element << ">\n";
	}

	else if(node->element=="vector" || node->element=="entity") {
		child << tabs << "<" << node->element << attr.str() << ">\n";
		std::list<Node *> & nodes= node->domain;
		for(std::list<Node *>::iterator it= nodes.begin(); it!=nodes.end(); ++it)
			child << write(*it, indent+1);
		child << tabs << "</" << node->element << ">\n";
	}

	else
		CHECK(false);

	return child.str();
}

void GuiUpgrade::clean() {
	std::list<Node>::iterator it= model.begin();
	for(++it; it!=model.end(); ++it)
		if(!it->owner)
			clean(&*it);
	for(--it; it!=model.begin(); --it)
		if(!it->owner)
			it= model.erase(it);
}

void GuiUpgrade::clean(Node * node) {
	std::list<Node *> nodes= node->domain;
	for(std::list<Node *>::iterator it= nodes.begin(); it!=nodes.end(); ++it) {
		remove(*it);
		clean(*it);
	}
}

GuiUpgrade::Node * GuiUpgrade::find(Node * node, const std::string & name) {
	if(node) {
		std::list<Node *> & nodes= node->domain;
		for(std::list<Node *>::iterator it= nodes.begin(); it!=nodes.end(); ++it)
			if((*it)->name==name)
				return *it;
	}
	return NULL;
}

void GuiUpgrade::remove(Node * node) {
	if(node) {
		CHECK(FindAndErase(node->owner->domain, node));
		node->owner= NULL;
	}
}

void GuiUpgrade::reown(Node * node, Node * owner) {
	if(node && owner) {
		remove(node);
		owner->domain.push_back(node);
		node->owner= owner;
	}
}

void GuiUpgrade::upgrade1() {
	std::list<Node *> elems;
	for(std::list<Node>::iterator it= model.begin(); it!=model.end(); ++it) {
		Node * node= &(*it);
		FindAndErase(node->flags, "SensitiveLink");
		FindAndErase(node->flags, "ModifyingLink");
		if(node->name=="name" && node->owner->type=="Gtk::ActionGroup")
			FindAndErase(node->flags, "Construct");
		if(node->type=="Gideon::UIElement")
			elems.push_back(node);
	}
	for(std::list<Node *>::iterator it= elems.begin(); it!=elems.end(); ++it) {
		Node * uielement= *it;
		Node * widget= find(uielement, "widget");
		Node * manager= find(uielement, "manager");
		Node * element= find(uielement, "element");

		if(manager)
			FindAndErase(manager->flags, "Construct");
		if(element)
			FindAndErase(element->flags, "Construct");

		std::string ui_element;
		std::vector<Node *> ui_definitions;

		std::string element_str= element ? element->text : "";
		Node * manager_node= manager ? find(&model.front(), manager->text) : NULL;
		if(manager_node) {
			Node * ui_defs= find(manager_node, "ui-definitions");
			if(ui_defs) {
				std::list<Node *> & defs= ui_defs->domain;
				for(std::list<Node *>::iterator it= defs.begin(); it!=defs.end(); ++it)
					ui_definitions.push_back(find(&model.front(), (*it)->text));
			}
		}

		for(int i=0; i<SIGNED(ui_definitions.size()); ++i) {
			Node * ui= find(ui_definitions[i], "ui");
			if(ui) {
				ui_element= UiElementExtract().extract(ui->cdata)[element_str];
				if(!ui_element.empty())
					break;
			}
		}

		if(ui_element=="menubar")
			uielement->type= "Gtk::MenuBar";
		else
			uielement->type= "Gtk::Toolbar";

		if(widget) {

			reown(find(widget, "border-width"), uielement);
			reown(find(widget, "events"), uielement);
			reown(find(widget, "extension-events"), uielement);
			reown(find(widget, "orientation"), uielement);
			reown(find(widget, "show-arrow"), uielement);
			reown(find(widget, "toolbar-style-set"), uielement);
			reown(find(widget, "toolbar-style"), uielement);
			reown(find(widget, "tooltips"), uielement);

			Node * wsignals= find(widget, "signals");
			if(wsignals) {
				Node * signals= find(uielement, "signals");
				if(!signals)
					reown(wsignals, uielement);
				else {
					int shift= signals->domain.size();
					std::list<Node *> nodes= wsignals->domain;
					for(std::list<Node *>::iterator it= nodes.begin(); it!=nodes.end(); ++it) {
						Node * node= *it;
						int index= FromString<int>(node->name);
						node->name= ToString(shift + index);
						reown(node, signals);
					}
				}
			}

			remove(widget);
		}
	}

	clean();
}

void GuiUpgrade::upgrade2() {
	for(std::list<Node>::iterator it= model.begin(); it!=model.end(); ++it) {
		Node * node= &(*it);
		bool found= FindAndErase(node->flags, "Translate");
		if(found)
			node->meta= "translate";
	}
}

void GuiUpgrade::upgrade4() {
	for(std::list<Node>::iterator it= model.begin(); it!=model.end(); ++it) {
		Node * node= &(*it);
		if(
			node->name=="position-set" && (
				node->owner->type=="Gtk::HPaned" ||
				node->owner->type=="Gtk::VPaned"
			)
		)
			FindAndErase(node->flags, "NoUndo");
	}
}

std::string GuiUpgrade::replaceColons(const std::string & str, const std::string & rep) {
	std::string result= str;
	int ch= result.find(':');
	if(ch != SIGNED(std::string::npos)) {
		CHECK(result[ch+1]==':');
		result.replace(ch, 2, rep);
	}
	return result;
}

void GuiUpgrade::upgrade5() {
	std::map<std::string, std::string> rename;
	std::set<std::string> enums;

	rename["Glib::ustring"] = "string";
	rename["Gideon::UniChar"] = "char";
	rename["Gtk::BuiltinIconSize"] = "GtkIconSize";

	enums.insert("GdkExtensionMode");
	enums.insert("GdkGravity");
	enums.insert("GdkWindowTypeHint");
	enums.insert("GtkShadowType");
	enums.insert("GtkPolicyType");
	enums.insert("GtkCornerType");
	enums.insert("GtkPositionType");
	enums.insert("GtkPackOptions");
	enums.insert("GtkButtonBoxStyle");
	enums.insert("GtkWindowType");
	enums.insert("GtkWindowPosition");
	enums.insert("GtkOrientation");
	enums.insert("GtkToolbarStyle");
	enums.insert("GtkResponseType");
	enums.insert("GtkFileChooserAction");
	enums.insert("GtkPackDirection");
	enums.insert("GtkJustification");
	enums.insert("GtkReliefStyle");
	enums.insert("GtkSpinButtonUpdatePolicy");
	enums.insert("GtkUpdateType");
	enums.insert("GtkProgressBarOrientation");
	enums.insert("GtkIconSize");
	enums.insert("GtkSizeGroupMode");
	enums.insert("GtkWrapMode");
	enums.insert("GtkSelectionMode");
	enums.insert("GtkArrowType");
	enums.insert("PangoEllipsizeMode");

	for(std::list<Node>::iterator it= model.begin(); it!=model.end(); ++it) {
		Node * node= &(*it);
		std::map<std::string, std::string>::iterator jt= rename.find(node->type);
		if(jt!=rename.end())
			node->type= jt->second;
		else
			node->type= replaceColons(node->type);
		if(enums.find(node->type)!=enums.end())
			node->text= Glib::ustring(replaceColons(node->text, "_")).uppercase();
	}
}

void GuiUpgrade::upgrade6() {
	for(std::list<Node>::iterator it= model.begin(); it!=model.end(); ++it) {
		Node * node= &(*it);
		node->flags.clear();
		if(node->owner && node->owner->element=="vector") {
			node->name= "";
			if(node->owner->name=="signals") {
				Emitter E= FromString<Emitter>(node->text);
				if(E.signame=="get-accessible")
					remove(node);
				else if(E.signame=="parent-changed") {
					E.signame= "parent-set";
					node->text= ToString(E);
				} else if(E.signame=="style-changed") {
					E.signame= "style-set";
					node->text= ToString(E);
				} else if(E.signame=="toolbar-style-changed") {
					E.signame= "style-changed";
					node->text= ToString(E);
				}
			}
		}
		if(node->type=="GtkPackOptions") {
			node->type= "GideonPackOptions";
			node->text.replace(0, 3, "GIDEON");
		}
	}
	clean();
}

void GuiUpgrade::upgrade8() {
	for(std::list<Node>::iterator it= model.begin(); it!=model.end(); ++it) {
		Node * node= &(*it);
		if(Begins(node->type, std::string("Gideon")))
			node->type.replace(0, 6, "Crow");
		if(node->type=="CrowPackOptions")
			node->text.replace(0, 6, "CROW");
	}
	for(std::list<Node>::iterator it= model.begin(); it!=model.end(); ++it) {
		Node * node= &(*it);
		if(
			node->type == "CrowBoxChild" && (
				node->owner->owner->type == "GtkHButtonBox" ||
				node->owner->owner->type == "GtkVButtonBox"
			)
		)
			node->type= "CrowButtonBoxChild";
	}
}

void GuiUpgrade::upgrade9() {
	std::list<Node *> elems;
	for(std::list<Node>::iterator it= model.begin(); it!=model.end(); ++it) {
		Node * node= &(*it);
		if(node->type=="GtkTooltips")
			elems.push_back(node);
		if(node->name=="window-placement")
			remove(node);
		if(node->name=="focus-child")
			remove(node);
		if(node->name=="focus-chain") {
			GType container= g_type_from_name(node->owner->type.c_str());
			if(g_type_is_a(container, GTK_TYPE_BIN))
				remove(node);
		}
	}
	for(std::list<Node *>::iterator it= elems.begin(); it!=elems.end(); ++it) {
		Node * ntooltips= *it;
		Node * ntips= find(ntooltips, "tips");
		Node * nenabled= find(ntooltips, "enabled");
		bool enabled= nenabled ? nenabled->text=="true" : true;
		if(ntips) {
			std::list<Node *> & tips= ntips->domain;
			for(std::list<Node *>::iterator it= tips.begin(); it!=tips.end(); ++it) {
				Node * nwidget= find(*it, "widget");
				Node * ntip= find(*it, "tip-text");
				if(nwidget && ntip && (!ntip->text.empty() || !ntip->meta.empty())) {
					Node * widget= find(&model.front(), nwidget->text);
					if(widget && !find(widget, "tooltip")) {
						model.push_back(Node());
						Node * ntooltip= &model.back();
						ntooltip->element= "scalar";
						ntooltip->name= "tooltip";
						ntooltip->type= "string";
						ntooltip->meta= ntip->meta;
						ntooltip->text= ntip->text;
						ntooltip->owner= widget;
						widget->domain.push_back(ntooltip);
						if(!enabled) {
							model.push_back(Node());
							Node * nmode= &model.back();
							nmode->element= "scalar";
							nmode->name= "tooltip-mode";
							nmode->type= "CrowTooltipMode";
							nmode->text= "CROW_TOOLTIP_NONE";
							nmode->owner= widget;
							widget->domain.push_back(nmode);
						}
					}
				}
			}
		}
		remove(ntooltips);
	}
	clean();
}

}
