/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2009 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#include "headers.h"
#include "forward.h"
#include "editors.h"

#include "palette.h"
#include "canvaswid.h"
#include "canvasuid.h"
#include "session.h"
#include "views2.h"

namespace Crow {

void RegisterEditors() {
	Palette * palette= Palette::get();
	palette->addEditor<PropertyEditorInput>("int");
	palette->addEditor<PropertyEditorInput>("unsigned");
	palette->addEditor<BoolPropertyEditor>("bool");
	palette->addEditor<PropertyEditorInput>("float");
	palette->addEditor<PropertyEditorInput>("double");
	palette->addEditor<PropertyEditorInput>("string");
	palette->addEditor<PropertyEditorInput>("char");
	palette->addEditor<WidgetCanvasEditor>("canvas|GtkWidget");
	palette->addEditor<ObjectPropertyEditor>("GlibObject");
	palette->addEditor<VectorPropertyEditor>("VECTOR");
	palette->addEditor<EnumPropertyEditor>("ENUM");
	palette->addEditor<FlagsPropertyEditor>("FLAGS");
	palette->addEditor<NullCanvasEditor>("canvas|GlibObject");
	palette->addEditor<UIDefinitionCanvasEditor>("canvas|CrowUIDefinition");
	palette->addEditor<ColorPropertyEditor>("GdkColor");
	palette->addEditor<PropertyEditorInput>("CrowPoint");
	palette->addEditor<EmitterPropertyEditor>("CrowEmitter");
	palette->addEditor<StockIDPropertyEditor>("GtkStockID");
	palette->addEditor<IconNamePropertyEditor>("icon-name|string");
	palette->addEditor<ElementPropertyEditor>("element|string");
	palette->addEditor<PropertyEditorInput>("GtkBorder");
}

// Editor ----------------------------------------------------------------------

void Editor::init(PSession sess) {
	session= sess;
	setEditorWidget(createEditorWidget());
}

void Editor::done() {
	session= PSession();
	if(widget) {
		widget->unsetEditor();
		widget= PEditorWidget();
	}
}

PEditorWidget Editor::getEditorWidget() {
	return widget;
}

Controller * Editor::getController() {
	return session->getManager()->getController();
}

SessionManager * Editor::getManager() {
	return session->getManager();
}

PSession Editor::getSession() {
	return session;
}

void Editor::setEditorWidget(PEditorWidget ew) {
	widget= ew;
	widget->setEditor(MakeRefPtr(this));
}

// HierarchyEditor -------------------------------------------------------------

void HierarchyEditor::created() {
	getEditorWidget()->getTreeView()->connect_selection_changed(sigc::mem_fun(*this, &HierarchyEditor::onTreeViewSelectionChanged));
	getEditorWidget()->getTreeView()->connect_context_menu(sigc::mem_fun(*static_cast<Editor *>(this), &HierarchyEditor::popupMenu));
	getManager()->connect_selection_changed(sigc::mem_fun(*this, &HierarchyEditor::onManagerSelectionChanged));
	getEditorWidget()->getTreeView()->setSetRowValues(sigc::mem_fun(*this, &HierarchyEditor::setRowValues));
}

void HierarchyEditor::load(bool reload) {

	Model * model= getManager()->getModel();
	const Nodes & nodes= *model->getRoot()->getOut();

	getEditorWidget()->getTreeView()->prepare();

	for(Nodes::const_iterator it= nodes.begin(); it!=nodes.end(); ++it) {
		PNode node= *it;
		getEditorWidget()->getTreeView()->update(model->getChildPath(node), node);
	}

	getEditorWidget()->getTreeView()->submit();

	onManagerSelectionChanged();

}

void HierarchyEditor::deleted() {
	getEditorWidget()->getTreeView()->prepare();
	getEditorWidget()->getTreeView()->submit();
}

void HierarchyEditor::setRowValues(PPolyelem elem) {

	PNode node= PNode::cast_static(elem->getObject());
	const std::string & type= *GetEntry(node->getType())->getTypeName();

	PPolycellText pcclass= NewRefPtr<PolycellText>();
	pcclass->setText(type);
	elem->setValue(0, pcclass);

	Glib::RefPtr<PolycellInput> pcname= NewRefPtr<PolycellInput>();
	pcname->setText(*node->getName());
	pcname->setAllowEditing(true);
	pcname->setAutoStop(true);
	pcname->connect_accepted(sigc::bind<PolycellInput *, PNode>(
		sigc::mem_fun(*this, &HierarchyEditor::onRename), GetPtr(pcname), node
	));
	pcname->setValidator(sigc::mem_fun(*this, &HierarchyEditor::validator));
	elem->setValue(1, pcname);

}

void HierarchyEditor::onManagerSelectionChanged() {
	const Nodes & nsel= *getManager()->getSelection();
	std::vector<PObject> selection(nsel.begin(), nsel.end());
	getEditorWidget()->getTreeView()->setSelection(selection, false);
}

void HierarchyEditor::onTreeViewSelectionChanged() {
	const std::vector<PObject> & selection= *getEditorWidget()->getTreeView()->getSelection();
	Nodes nsel;
	std::transform(selection.begin(), selection.end(), std::back_inserter(nsel), &PNode::cast_static<Object>);
	getManager()->submitSelection(nsel);
}

bool HierarchyEditor::validator(const Glib::ustring & oldtext, const Glib::ustring & newtext) {
	return getManager()->getModel()->validName(newtext);
}

void HierarchyEditor::onRename(PolycellInput * polycell, PNode node) {
	std::string name= *polycell->getText();
	getManager()->begin();
	getManager()->getModel()->setName(node, name);
	getManager()->commit();
}

// NullCanvasEditor ------------------------------------------------------------

PEditorWidget NullCanvasEditor::createEditorWidget() {
	PEditorWidget widget= NewRefPtr<EditorWidget>();
	Gtk::Frame * frame= Gtk::manage(new Gtk::Frame());
	frame->set_shadow_type(Gtk::SHADOW_IN);
	widget->add(*frame);
	frame->show();
	return widget;
}

// PropertyEditor --------------------------------------------------------------

PPropertyEditorWidget PropertyEditor::getEditorWidget() {
	return PPropertyEditorWidget::cast_static(Editor::getEditorWidget());
}

void PropertyEditor::created() {
	getEditorWidget()->connect_accepted(sigc::mem_fun(*this, &PropertyEditor::onAccepted));
}

TypeId PropertyEditor::getEditingFinalType() {
	return getSession()->getFinalType();
}

Property * PropertyEditor::getProperty1() {
	return getSession()->getProperty1();
}

Any PropertyEditor::getScalar() {
	return getSession()->getScalar();
}

void PropertyEditor::setScalar(Any value) {
	getManager()->begin();
	getSession()->setScalarOnly(value);
	getManager()->commit();
}

void PropertyEditor::setScalarAndMeta(Any value, const Glib::ustring & meta) {
	getManager()->begin();
	getSession()->setScalarAndMeta(value, meta);
	getManager()->commit();
}

// VectorPropertyEditor --------------------------------------------------------

void VectorPropertyEditor::load(bool reload) {
	if(getSession()->isFuzzy())
		getEditorWidget()->setFuzzy();
	else {
		TypeId type= getSession()->getVectorType();
		int size= -1;
		if(!getSession()->isMultiple()) {
			PNode node= getSession()->getNode1();
			size= node->getOut()->size();
		}
		std::string number;
		if(size==-1)
			number= "...";
		else if(size!=0)
			number= ToString(size);
		std::string str= *GetEntry(type)->getTypeName() + "[" + number + "]";
		getEditorWidget()->setText(str);
	}
}

// ObjectPropertyEditor --------------------------------------------------------

void ObjectPropertyEditor::created() {
	getEditorWidget()->setPopupEnabled(true);
	getEditorWidget()->setEntryEditable(false);
	getEditorWidget()->connect_popup(sigc::mem_fun(*this, &ObjectPropertyEditor::onPopup));
}

void ObjectPropertyEditor::load(bool reload) {
	PNode target= getSession()->getLinkTarget();
	if(target) {
		getEditorWidget()->setText(*target->getName());
		return;
	}
	TypeId type= getSession()->getEntityType();
	if(type) {
		getEditorWidget()->setText(*GetEntry(type)->getTypeName());
		return;
	}
	if(getSession()->areEmptyLinks()) {
		getEditorWidget()->setText("NULL");
		return;
	}
	getEditorWidget()->setFuzzy();
}

void ObjectPropertyEditor::onPopup() {

	Gtk::Menu * menu= Gtk::manage(new Gtk::Menu());
	menu->items().push_back(Gtk::Menu_Helpers::MenuElem("Cancel"));
	menu->items().push_back(Gtk::Menu_Helpers::SeparatorElem());

	Model * model= getManager()->getModel();
	TypeId baseType= getSession()->getObjectBaseType();

	menu->items().push_back(Gtk::Menu_Helpers::MenuElem("Clear", sigc::mem_fun(*this, &ObjectPropertyEditor::onMenuClear)));
	if(!getSession()->canDefault())
		menu->items().back().get_child()->set_sensitive(false);

	Nodes objects;
	if(baseType) {
		if(getSession()->areWeakLinks()) {
			objects= model->query(baseType, false);
			PNode target= getSession()->getLinkTarget();
			if(target)
				FindAndErase(objects, target);
			for(Nodes::iterator it= objects.begin(); it!=objects.end(); )
				if(
					!GetEntry<EntityEntry>(baseType)->hasFlag(efTopMaster) &&
					GetEntry<EntityEntry>((*it)->getType())->hasFlag(efTopMaster) &&
					!getSession()->allowTopMasters()
				)
					it= objects.erase(it);
				else
					++it;
		} else if(!getSession()->isMultiple()) {
			objects= model->query(baseType, false);
			PNode object= model->raiseOwner(getSession()->getNode1());
			FindAndErase(objects, object);
			for(Nodes::iterator it= objects.begin(); it!=objects.end(); )
				if(
					model->findMaster(*it) || model->hasMaster(object, *it) ||
					GetEntry<EntityEntry>((*it)->getType())->hasFlag(efTopMaster)
				)
					it= objects.erase(it);
				else
					++it;
		}
	}

	if(!objects.empty()) {
		SortByName(&objects);
		Gtk::Menu * selectMenu= Gtk::manage(new Gtk::Menu());
		for(Nodes::iterator it=objects.begin(); it!=objects.end(); ++it) {
			selectMenu->items().push_back(Gtk::Menu_Helpers::MenuElem(
				Underscores(*(*it)->getName()),
				sigc::bind<PNode>(
					sigc::mem_fun(*this, &ObjectPropertyEditor::onMenuSelect), *it)
			));
		}
		menu->items().push_back(Gtk::Menu_Helpers::MenuElem("Select", *selectMenu));
	} else {
		menu->items().push_back(Gtk::Menu_Helpers::MenuElem("Select"));
		menu->items().back().get_child()->set_sensitive(false);
	}

	bool create= false;
	if(baseType && !getSession()->hasLinkOnly()) {
		bool show_dw= getController()->getContext()->getKeyValue<bool>("show_dw");
		std::vector<TypeId> types= Palette::get()->queryTypes(baseType);
		if(!types.empty()) {
			Gtk::Menu * createMenu= Gtk::manage(new Gtk::Menu());
			for(int i=0; i<SIGNED(types.size()); ++i) {
				EntityEntry * entry= GetEntry<EntityEntry>(types[i]);
				if(
					!entry->hasFlag(efAbstract) &&
					!entry->hasFlag(efInternal) &&
					!entry->hasFlag(efTopMaster) &&
					(show_dw || !entry->hasFlag(efDebug))
				)
					createMenu->items().push_back(Gtk::Menu_Helpers::MenuElem(
						Underscores(*entry->getTypeName()),
						sigc::bind<TypeId>(
							sigc::mem_fun(*this, &ObjectPropertyEditor::onMenuCreate), types[i])
					));
			}
			menu->items().push_back(Gtk::Menu_Helpers::MenuElem("Create", *createMenu));
			create= true;
		}
	}
	if(!create) {
		menu->items().push_back(Gtk::Menu_Helpers::MenuElem("Create"));
		menu->items().back().get_child()->set_sensitive(false);
	}

	menu->popup(0, GDK_CURRENT_TIME);
}

void ObjectPropertyEditor::onMenuClear() {
	getManager()->begin();
	getSession()->resetObject();
	getManager()->commit();
}

void ObjectPropertyEditor::onMenuSelect(PNode object) {
	getManager()->begin();
	getSession()->setObjectAsLink(object);
	getManager()->commit();
}

void ObjectPropertyEditor::onMenuCreate(TypeId type) {
	getManager()->begin();
	getSession()->setObjectAsEntity(type);
	getManager()->commit();
}

// PropertyEditorInput ---------------------------------------------------------

void PropertyEditorInput::created() {
	PropertyEditor::created();
	if(getSession()->areTranslatable()) {
		getEditorWidget()->setPopupEnabled(true);
		getEditorWidget()->connect_popup(sigc::mem_fun(*this, &PropertyEditorInput::onPopup));
	}
}

void PropertyEditorInput::load(bool reload) {
	Any value= getScalar();
	if(value)
		getEditorWidget()->setText(toText(value));
	else
		getEditorWidget()->setFuzzy();
}

void PropertyEditorInput::onAccepted() {
	Glib::ustring str= *getEditorWidget()->getText();
	Any value;
	try {
		value= fromText(str);
	} catch(const RuntimeError &) {
		load(true);
		return;
	}
	setScalar(value);
}

Glib::ustring PropertyEditorInput::toText(Any value) {
	return value->toString();
}

Any PropertyEditorInput::fromText(const Glib::ustring & text) {
	return CAny::createFromString(getEditingFinalType(), text);
}

void PropertyEditorInput::onPopup() {
	Glib::ustring text= getEditorWidget()->getEntryText();
	Translatable meta= Translatable(getSession()->getMeta());
	for(;;) {
		TextInputDialog dialog;
		dialog.set_transient_for(*getController()->getContext()->getWindow());
		dialog.setText(text);
		dialog.setMeta(meta);
		if(getController()->getContext()->runDialog(&dialog)==Gtk::RESPONSE_OK) {
			text= dialog.getText();
			meta= dialog.getMeta();
			if(!meta.isValid()) {
				getController()->getContext()->showMessage("Error: Prefix and/or Comments contain '|' and/or '*/'");
				continue;
			}
			setScalarAndMeta(CAny::createString(text), meta.getMeta());
		}
		break;
	}
}

// PropertyEditorChoice --------------------------------------------------------

void PropertyEditorChoice::load(bool reload) {
	Any value= getScalar();
	if(value)
		getEditorWidget()->setText(value->toString());
	else
		getEditorWidget()->setFuzzy();
}

// BoolPropertyEditor ----------------------------------------------------------

void BoolPropertyEditor::created() {
	PropertyEditorChoice::created();
	getEditorWidget()->addChoice("false");
	getEditorWidget()->addChoice("true");
}

void BoolPropertyEditor::onAccepted() {
	std::string str= *getEditorWidget()->getText();
	Any value= CAny::createBool(FromString<bool>(str));
	setScalar(value);
}

// EnumPropertyEditor ----------------------------------------------------------

void EnumPropertyEditor::created() {
	PropertyEditorChoice::created();
	getEditorWidget()->setEllipsize(Pango::ELLIPSIZE_START);
	EnumEntry * entry= GetEntry<EnumEntry>(getEditingFinalType());
	const EnumEntry::Values & values= *entry->getValues();
	for(EnumEntry::Values::const_iterator it= values.begin(); it!=values.end(); ++it)
		getEditorWidget()->addChoice((*it).first);
}

void EnumPropertyEditor::onAccepted() {
	std::string str= *getEditorWidget()->getText();
	EnumEntry * entry= GetEntry<EnumEntry>(getEditingFinalType());
	Any value= CAny::createEnum(getEditingFinalType(), entry->getByName(str));
	setScalar(value);
}

// FlagsPropertyEditor ---------------------------------------------------------

void FlagsPropertyEditor::created() {
	PropertyEditorInput::created();
	getEditorWidget()->setPopupEnabled(true);
	getEditorWidget()->connect_popup(sigc::mem_fun(*this, &FlagsPropertyEditor::onPopup));
}

void FlagsPropertyEditor::onPopup() {
	FlagsDialog dialog;
	dialog.set_transient_for(*getController()->getContext()->getWindow());

	FlagsEntry * entry= GetEntry<FlagsEntry>(getEditingFinalType());
	dialog.set_title(*entry->getTypeName());
	dialog.setOptions(*entry->getValues());

	Any current;
	try {
		current= entry->createFromString(getEditorWidget()->getEntryText());
	} catch(const RuntimeError &) {
	}
	if(current)
		dialog.setValue(current->getFlags());

	if(getController()->getContext()->runDialog(&dialog)==Gtk::RESPONSE_OK) {
		Bits value= dialog.getValue();
		CHECK(entry->valid(value));
		setScalar(CAny::createFlags(getEditingFinalType(), value));
	}

}

// EmitterPropertyEditor -------------------------------------------------------

EmitterPropertyEditor::EmitterPropertyEditor() {
	lock= false;
}

void EmitterPropertyEditor::created() {
	PropertyEditorInput::created();
	getEditorWidget()->setValidator(sigc::mem_fun(*this, &EmitterPropertyEditor::validator));
	getEditorWidget()->setPopupEnabled(true);
	getEditorWidget()->connect_popup(sigc::mem_fun(*this, &EmitterPropertyEditor::onPopup));
}

std::string EmitterPropertyEditor::getMySignature() {
	const std::string & signame= getScalar()->get<Emitter>()->signame;
	PNode owner= getSession()->getNode1()->getOwner()->getOwner();
	PGlibObjectView view= PGlibObjectView::cast_dynamic(getController()->getView(owner));
	std::string signature= view->findSignal(signame);
	return signature;
}

bool EmitterPropertyEditor::validator(const Glib::ustring & oldtext, const Glib::ustring & newtext) {
	if(newtext.empty())
		return true;
	if(!ValidName(newtext))
		return false;
	Controller::Emitters emitters= getController()->getEmitters();
	Controller::Emitters::iterator it= emitters.find(newtext);
	if(it==emitters.end())
		return true;
	std::string signature= it->second.signature;
	return signature == getMySignature();
}

Glib::ustring EmitterPropertyEditor::toText(Any value) {
	return value->get<Emitter>()->mapname;
}

Any EmitterPropertyEditor::fromText(const Glib::ustring & text) {
	Emitter emitter= *getScalar()->get<Emitter>();
	emitter.mapname= text;
	return CAny::createEmitter(emitter);
}

void EmitterPropertyEditor::onPopup() {
	Gtk::Menu * menu= Gtk::manage(new Gtk::Menu());
	menu->items().push_back(Gtk::Menu_Helpers::MenuElem("Cancel"));
	menu->items().push_back(Gtk::Menu_Helpers::SeparatorElem());

	Gtk::RadioMenuItem::Group group;
	menu->items().push_back(Gtk::Menu_Helpers::RadioMenuElem(group, "after"));
	menu->items().push_back(Gtk::Menu_Helpers::RadioMenuElem(
		group, "before",
		sigc::bind<Gtk::Menu *>(sigc::mem_fun(*this, &EmitterPropertyEditor::onOrderSelect), menu)
	));

	lock= true;
	dynamic_cast<Gtk::RadioMenuItem *>(&menu->items()[3])->set_active(!getScalar()->get<Emitter>()->after);
	lock= false;

	std::set<std::string> choices;
	Controller::Emitters emitters= getController()->getEmitters();
	std::string mysign= getMySignature();
	for(Controller::Emitters::iterator it= emitters.begin(); it!=emitters.end(); ++it) {
		std::string signature= it->second.signature;
		if(signature==mysign && !it->first.empty())
			choices.insert(it->first);
	}

	choices.erase(getScalar()->get<Emitter>()->mapname);

	if(!choices.empty()) {
		menu->items().push_back(Gtk::Menu_Helpers::SeparatorElem());
		for(std::set<std::string>::iterator it= choices.begin(); it!=choices.end(); ++it)
			menu->items().push_back(Gtk::Menu_Helpers::MenuElem(
				Underscores(*it),
				sigc::bind<std::string>(sigc::mem_fun(*this, &EmitterPropertyEditor::onMenuSelect), *it)
			));
	}

	menu->popup(0, GDK_CURRENT_TIME);
}

void EmitterPropertyEditor::onOrderSelect(Gtk::Menu * menu) {
	if(!lock) {
		bool after= dynamic_cast<Gtk::RadioMenuItem *>(&menu->items()[2])->get_active();
		Emitter emitter= *getScalar()->get<Emitter>();
		emitter.after= after;
		setScalar(CAny::createEmitter(emitter));
	}
}

void EmitterPropertyEditor::onMenuSelect(const std::string & name) {
	Emitter emitter= *getScalar()->get<Emitter>();
	emitter.mapname= name;
	setScalar(CAny::createEmitter(emitter));
}

// ColorPropertyEditor ---------------------------------------------------------

void ColorPropertyEditor::load(bool reload) {
	Any value= getScalar();
	if(value)
		getEditorWidget()->setColor(*value->get<Gdk::Color>());
	else
		getEditorWidget()->setFuzzy();
}

void ColorPropertyEditor::onAccepted() {
	const Gdk::Color & color= *getEditorWidget()->getColor();
	Any value= CAny::create(getEditingFinalType(), color);
	setScalar(value);
}

// StockIDPropertyEditor -------------------------------------------------------

void StockIDPropertyEditor::created() {
	PropertyEditorInput::created();
	getEditorWidget()->setPopupEnabled(true);
	getEditorWidget()->connect_popup(sigc::mem_fun(*this, &StockIDPropertyEditor::onPopup));
}

void StockIDPropertyEditor::onPopup() {
	IconSelectDialog dialog;
	dialog.set_title("Stock Item Select");
	dialog.set_transient_for(*getController()->getContext()->getWindow());

	std::vector<Gtk::StockID> ids= Gtk::Stock::get_ids();
	std::vector<std::pair<PGdkPixbuf, Glib::ustring> > choices;
	choices.reserve(ids.size());
	for(int i=0; i<SIGNED(ids.size()); ++i) {
		PGdkPixbuf pixbuf= dialog.render_icon(ids[i], Gtk::ICON_SIZE_MENU);
		choices.push_back(std::make_pair(pixbuf, ids[i].get_string()));
	}

	dialog.setChoices(choices);
	dialog.setChoice(getEditorWidget()->getEntryText());

	if(getController()->getContext()->runDialog(&dialog) == Gtk::RESPONSE_OK)
		setScalar(CAny::create(FindType("GtkStockID"), Gtk::StockID(dialog.getChoice())));
}

// IconNamePropertyEditor ------------------------------------------------------

void IconNamePropertyEditor::created() {
	PropertyEditorInput::created();
	getEditorWidget()->setPopupEnabled(true);
	getEditorWidget()->connect_popup(sigc::mem_fun(*this, &IconNamePropertyEditor::onPopup));
}

void IconNamePropertyEditor::onPopup() {
	IconSelectDialog dialog;
	dialog.set_transient_for(*getController()->getContext()->getWindow());

	std::vector<std::pair<PGdkPixbuf, Glib::ustring> > choices;

	GList * list= gtk_icon_theme_list_icons(gtk_icon_theme_get_default(), NULL);
	if(list) {
		choices.reserve(g_list_length(list));
		for(GList * elem= list; elem!=NULL; elem= g_list_next(elem)) {
			char * name= static_cast<char *>(elem->data);

			PGdkPixbuf pixbuf;
			try {
				pixbuf= Gtk::IconTheme::get_default()->load_icon(name, 16, Gtk::IconLookupFlags());
			} catch(const Glib::Error &) {
			}

			if(pixbuf && pixbuf->get_width()==16 && pixbuf->get_height()==16)
				choices.push_back(std::make_pair(pixbuf, name));

			g_free(name);
		}
		g_list_free(list);
	}

	dialog.setChoices(choices);
	dialog.setChoice(getEditorWidget()->getEntryText());

	if(getController()->getContext()->runDialog(&dialog)==Gtk::RESPONSE_OK)
		setScalar(CAny::createString(dialog.getChoice()));

}

// ElementPropertyEditor -------------------------------------------------------

void ElementPropertyEditor::created() {
	PropertyEditorInput::created();
	if(!getSession()->isMultiple())
		getEditorWidget()->setPopupEnabled(true);
	getEditorWidget()->connect_popup(sigc::mem_fun(*this, &ElementPropertyEditor::onPopup));
}

void ElementPropertyEditor::onPopup() {
	Gtk::Menu * menu= Gtk::manage(new Gtk::Menu());
	menu->items().push_back(Gtk::Menu_Helpers::MenuElem("Cancel"));

	Model * model= getManager()->getModel();
	PNode owner= getSession()->getNode1()->getOwner();
	PNode manager= model->find(owner, "manager")->getLink();
	std::set<std::string> choices;
	if(manager) {

		PNode definitions= model->find(manager, "ui-definitions");
		const Nodes & out= *definitions->getOut();
		for(Nodes::const_iterator it= out.begin(); it!=out.end(); ++it) {
			PNode definition= (*it)->getLink();
			if(definition) {
				const Glib::ustring & ui= *model->find(definition, "ui")->getValue()->getString();
				std::map<std::string, std::string> elements= UiElementExtract().extract(ui);
				for(std::map<std::string, std::string>::iterator it= elements.begin(); it!=elements.end(); ++it)
					if(
						(it->second=="menubar" && *Palette::get()->getEntry(owner->getType())->getTypeName()=="GtkMenuBar") ||
						(it->second=="toolbar" && *Palette::get()->getEntry(owner->getType())->getTypeName()=="GtkToolbar")
					)
						choices.insert(it->first);
			}
		}

		const Nodes & links= *manager->getIn();
		for(Nodes::const_iterator it= links.begin(); it!=links.end(); ++it) {
			PNode obj= model->raiseOwner(*it);
			if(Glib::RefPtr<UIElementView>::cast_dynamic(getManager()->getController()->getView(obj))) {
				PNode elem= model->find(obj, "element");
				choices.erase(*elem->getValue()->getString());
			}
		}

	}

	if(!choices.empty()) {
		menu->items().push_back(Gtk::Menu_Helpers::SeparatorElem());
		for(std::set<std::string>::iterator it= choices.begin(); it!=choices.end(); ++it)
			menu->items().push_back(Gtk::Menu_Helpers::MenuElem(
				Underscores(*it),
				sigc::bind<std::string>(sigc::mem_fun(*this, &ElementPropertyEditor::onMenuSelect), *it)
			));
	}

	menu->popup(0, GDK_CURRENT_TIME);
}

void ElementPropertyEditor::onMenuSelect(const std::string & choice) {
	setScalar(CAny::createString(choice));
}

}
