/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2009 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#include "headers.h"
#include "forward.h"
#include "model.h"

#include "palette.h"

namespace Crow {

// Node ------------------------------------------------------------------------

Any Node::getValue() {
	CHECK(role==nrScalar);
	return value;
}

bool Node::isInactive() {
	return !owner;
}

PNode Node::getOwner() {
	return owner;
}

PNode Node::getLink() {
	CHECK(role==nrLink);
	return link;
}

Node::Node(NodeRole r, TypeId t, StateFlags s) {
	role= r;
	type= t;
	state= s;
	modified= false;
}

PNode Node::create(NodeRole r, TypeId t, StateFlags s) {
	CHECK((r!=nrVector && r!=nrLink) || !t);
	CHECK((r!=nrVector && r!=nrEntity) || s==sfNone);
	return NewRefPtr(new Node(r, t, s));
}

void Node::setValue(Any any) {
	CHECK(role==nrScalar);
	value= any;
}

void Node::pushOut(PNode node) {
	CHECK(role!=nrLink && role!=nrScalar);
	CHECK(PushIfAbsent(out, node));
}

void Node::pushIn(PNode node) {
	CHECK(node->getRole()==nrLink);
	CHECK(PushIfAbsent(in, node));
}

void Node::eraseOut(PNode node) {
	CHECK(role!=nrLink && role!=nrScalar);
	CHECK(FindAndErase(out, node));
}

void Node::eraseIn(PNode node) {
	CHECK(node->getRole()==nrLink);
	CHECK(FindAndErase(in, node));
}

void Node::setOwner(PNode nowner) {
	CHECK(owner ^ nowner);
	owner= nowner;
}

void Node::setLink(PNode nlink) {
	CHECK(role==nrLink);
	link= nlink;
}

// Operation -------------------------------------------------------------------

POperation Operation::create(OperationCode oc, PNode node) {
	POperation op= NewRefPtr<Operation>();
	op->code= oc;
	op->node= node;
	return op;
}

POperation Operation::create(OperationCode oc, PNode node, const std::string & oldname, const std::string & newname) {
	POperationName op= NewRefPtr<OperationName>();
	op->code= oc;
	op->node= node;
	op->oldname= oldname;
	op->newname= newname;
	return op;
}

POperation Operation::create(OperationCode oc, PNode node, const Glib::ustring & oldmeta, const Glib::ustring & newmeta) {
	POperationMeta op= NewRefPtr<OperationMeta>();
	op->code= oc;
	op->node= node;
	op->oldmeta= oldmeta;
	op->newmeta= newmeta;
	return op;
}

POperation Operation::create(OperationCode oc, PNode node, Any oldvalue, Any newvalue) {
	POperationValue op= NewRefPtr<OperationValue>();
	op->code= oc;
	op->node= node;
	op->oldvalue= oldvalue;
	op->newvalue= newvalue;
	return op;
}

POperation Operation::create(OperationCode oc, PNode node, PNode link) {
	POperationLink op= NewRefPtr<OperationLink>();
	op->code= oc;
	op->node= node;
	op->link= link;
	return op;
}

POperation Operation::create(OperationCode oc, PNode node, PNode oldnode, PNode newnode) {
	POperationNode op= NewRefPtr<OperationNode>();
	op->code= oc;
	op->node= node;
	op->oldnode= oldnode;
	op->newnode= newnode;
	return op;
}

// Model -----------------------------------------------------------------------

Model::Model() {
	root= Node::create(nrNone, InvalidTypeId);
	dirty= false;
	readOnly= false;
	gatherLinks= false;
	mode= umNone;
	changeType= ctNone;
	clearHistory();
}

Model::~Model() {
	CHECK(empty());
	CHECK(emptyHistory());
}

PNode Model::getRoot() {
	return root;
}

bool Model::empty() {
	return root->getOut()->empty();
}

bool Model::emptyHistory() {
	if(history.empty()) {
		CHECK(pointer==history.end());
		CHECK(hpointer==0);
		CHECK(hsize==0);
		return true;
	} else
		return false;
}

bool Model::testChange(ChangeType ctype) {
	return (changeType & ctype) == changeType;
}

bool Model::validName(const std::string & name) {
	return ValidName(name) && !find(root, name);
}

std::string Model::makeUnique(const std::string & name) {

	PNode owner= root;

	if(!find(owner, name))
		return name;

	std::string str;
	int num;
	SplitNumber(name, &str, &num);

	int count= num == -1 ? 0 : num;
	std::string unique;

	do {
		++count;
		unique= str + ToString(count);
	} while(find(owner, unique));

	return unique;
}

std::string Model::suggestName(TypeId type) {
	std::string name= Palette::get()->getEntry(type)->suggestName();
	return makeUnique(name);
}

bool Model::isModified() {
	return root->isModified();
}

void Model::tagModified(PNode node) {
	setModified(node ? node : root);
}

IdPath Model::getNodePath(PNode node, PNode cut) {

	IdPath path;
	PNode N= node;
	PNode stop= cut ? cut : root;

	while(N != stop) {
		path.insert(path.begin(), *N->getName());
		N= N->getOwner();
	}

	return path;
}

IdPath Model::getChildPath(PNode node) {
	IdPath path;
	PNode N= node;
	for(;;) {
		path.insert(path.begin(), *N->getName());
		PNode parent= findParent(N);
		if(!parent)
			return path;
		N= parent;
	}
}

Nodes Model::getNested(PNode node, bool all) {
	Nodes result;
	getNested(node, &result, all);
	return result;
}

Nodes Model::getChildren(PNode node, bool all) {
	Nodes result;
	getChildren(node, &result, all);
	return result;
}

Nodes Model::getSlaves(PNode node, bool all) {
	Nodes result;
	getSlaves(node, &result, all);
	return result;
}

NodeSet Model::growSlaves(const Nodes & nodes) {
	NodeSet result;
	for(Nodes::const_iterator it= nodes.begin(); it!=nodes.end(); ++it) {
		result.insert(*it);
		Nodes slaves= getSlaves(*it);
		result.insert(slaves.begin(), slaves.end());
	}
	return result;
}

Nodes Model::selectMasters(const Nodes & nodes) {
	Nodes sel= nodes;
	for(Nodes::iterator it=sel.begin(); it!=sel.end(); ++it)
		for(Nodes::iterator jt=sel.begin(); jt!=sel.end(); ++jt)
			if(it!=jt && *jt) {
				if(hasMaster(*it, *jt)) {
					*it= PNode();
					break;
				}
			}
	Nodes result;
	for(Nodes::iterator it=sel.begin(); it!=sel.end(); ++it)
		if(*it)
			result.push_back(*it);
	return result;
}

Nodes Model::selectParents(const Nodes & nodes) {
	Nodes sel= nodes;
	for(Nodes::iterator it=sel.begin(); it!=sel.end(); ++it)
		for(Nodes::iterator jt=sel.begin(); jt!=sel.end(); ++jt)
			if(it!=jt && *jt) {
				if(hasParent(*it, *jt)) {
					*it= PNode();
					break;
				}
			}
	Nodes result;
	for(Nodes::iterator it=sel.begin(); it!=sel.end(); ++it)
		if(*it)
			result.push_back(*it);
	return result;
}

Nodes Model::query(TypeId baseType, bool all) {
	Nodes result;
	query(root, baseType, &result, all);
	return result;
}

PNode Model::findObject(const std::string & name) {
	return find(root, name);
}

PNode Model::find(PNode owner, const std::string & name) {
	const Nodes & out= *owner->getOut();
	for(Nodes::const_iterator it= out.begin(); it!=out.end(); ++it)
		if(*(*it)->getName() == name)
			return *it;
	return PNode();
}

PNode Model::findParent(PNode node) {
	PNode parent;
	PNode child= findChildNode(node);
	if(child)
		parent= child->getOwner()->getOwner();
	return parent;
}

PNode Model::findMaster(PNode node) {
	const Nodes & links= *node->getIn();
	for(Nodes::const_iterator it= links.begin(); it!=links.end(); ++it) {
		PNode link= *it;
		if(!link->isWeakLink()) {
			PNode owner= raiseOwner(link);
			CHECK(owner && owner->getRole()==nrEntity);
			return owner;
		}
	}
	return PNode();
}

PNode Model::raiseOwner(PNode node) {
	PNode N= node;
	for(;;) {
		PNode owner= N->getOwner();
		if(owner==root)
			return N;
		N= owner;
	}
}

PNode Model::raiseParent(PNode node) {
	PNode N= node;
	for(;;) {
		PNode parent= findParent(N);
		if(!parent)
			return N;
		N= parent;
	}
}

PNode Model::raiseMaster(PNode node) {
	PNode N= node;
	for(;;) {
		PNode master= findMaster(N);
		if(!master)
			return N;
		N= master;
	}
}

bool Model::hasParent(PNode child, PNode theParent) {
	CHECK(theParent);
	PNode N= child;
	for(;;) {
		PNode parent= findParent(N);
		if(!parent)
			return false;
		if(parent==theParent)
			return true;
		N= parent;
	}
}

bool Model::hasMaster(PNode child, PNode theMaster) {
	CHECK(theMaster);
	PNode N= child;
	for(;;) {
		PNode master= findMaster(N);
		if(!master)
			return false;
		if(master==theMaster)
			return true;
		N= master;
	}
}

PNode Model::findChildNode(PNode node) {
	const Nodes & links= *node->getIn();
	for(Nodes::const_iterator it= links.begin(); it!=links.end(); ++it) {
		PNode link= *it;
		if(*link->getName()=="widget") {
			PNode child= link->getOwner();
			if(child && isChildrenVector(child->getOwner()))
				return child;
		}
	}
	return PNode();
}

PNode Model::getWidgetNode(PNode node) {
	CHECK(node);
	PNode link= find(node, "widget");
	CHECK(link);
	PNode widget= link->getLink();
	CHECK(widget);
	return widget;
}

Any Model::getScalar(PNode owner, const std::string & name) {
	PNode node= find(owner, name);
	return node ? node->getValue() : Any();
}

bool Model::vectorEmptyOrNullLinks(PNode node) {
	CHECK(node->getRole()==nrVector);
	const Nodes & out= *node->getOut();
	for(Nodes::const_iterator it= out.begin(); it!=out.end(); ++it) {
		if((*it)->getRole()==nrLink && !(*it)->getLink())
			continue;
		return false;
	}
	return true;
}

PNode Model::createScalar(PNode owner, const std::string & name, Any value, StateFlags sf) {
	CHECK(value && value->getType());
	PNode node= createScalar(owner, value->getType(), name, sf);
	op_setValue(node, value);
	return node;
}

PNode Model::createScalar(PNode owner, TypeId type, const std::string & name, StateFlags sf) {
	CHECK(type);
	CHECK(!find(owner, name));
	PNode node= createNode(nrScalar, type, sf);
	op_setOwner(node, owner);
	op_setName(node, name);
	op_pushOut(owner, node);
	setModified(node);
	return node;
}

PNode Model::createVector(PNode owner, const std::string & name) {
	CHECK(!find(owner, name));
	PNode node= createNode(nrVector, InvalidTypeId);
	op_setOwner(node, owner);
	op_setName(node, name);
	op_pushOut(owner, node);
	setModified(node);
	return node;
}

PNode Model::createEntity(PNode owner, TypeId type, const std::string & name) {
	CHECK(!find(owner, name));
	PNode node= createNode(nrEntity, type);
	op_setOwner(node, owner);
	op_setName(node, name);
	op_pushOut(owner, node);
	setModified(node);
	return node;
}

PNode Model::createLink(PNode owner, const std::string & name, StateFlags sf) {
	CHECK(!find(owner, name));
	PNode node= createNode(nrLink, InvalidTypeId, sf);
	op_setOwner(node, owner);
	op_setName(node, name);
	op_pushOut(owner, node);
	setModified(node);
	return node;
}

PNode Model::vectorCreateEntity(PNode owner, TypeId type) {
	int index= owner->getOut()->size();
	return createEntity(owner, type, ToString(index));
}

PNode Model::vectorCreateScalar(PNode owner, Any value, StateFlags sf) {
	int index= owner->getOut()->size();
	return createScalar(owner, ToString(index), value, sf);
}

PNode Model::vectorCreateLink(PNode owner, StateFlags sf) {
	int index= owner->getOut()->size();
	return createLink(owner, ToString(index), sf);
}

void Model::cut(PNode node) {
	PNode owner= node->getOwner();
	op_eraseOut(owner, node);
	op_setOwner(node, PNode());
	CHECK(node->getIn()->empty());
	setModified(owner);
}

void Model::paste(PNode node, PNode owner) {
	CHECK(!find(owner, *node->getName()));
	op_pushOut(owner, node);
	op_setOwner(node, owner);
	setModified(node);
}

void Model::remove(PNode node) {
	CHECK(node && !node->isInactive());

	PNode owner= node->getOwner();

	if(node->getRole()==nrLink) {

		if(PNode target= node->getLink()) {
			if(!node->isWeakLink())
				remove(target);
			else
				clearLink(node);
		}
		CHECK(!node->getLink());

	} else if(node->getRole()!=nrScalar) {

		while(!node->getOut()->empty())
			remove(node->getOut()->front());
		CHECK(node->getOut()->empty());

	}

	while(!node->getIn()->empty())
		clearLink(node->getIn()->front());
	CHECK(node->getIn()->empty());

	setModified(owner);
	op_eraseOut(owner, node);

	op_setOwner(node, PNode());

}

void Model::rootRemove(PNode entity) {
	PNode child= findChildNode(entity);
	if(child)
		vectorRemove(child);
	else
		remove(entity);
}

void Model::vectorClear(PNode node) {
	while(!node->getOut()->empty())
		remove(node->getOut()->front());
}

void Model::vectorRemove(PNode node) {
	int index= FromString<int>(*node->getName());
	PNode owner= node->getOwner();
	remove(node);
	const Nodes & out= *owner->getOut();
	for(Nodes::const_iterator it= out.begin(); it!=out.end(); ++it) {
		int i= FromString<int>(*(*it)->getName());
		if(i > index)
			setName(*it, ToString(i-1));
	}
}

void Model::vectorMove(PNode node, int shift) {
	int index= FromString<int>(*node->getName());
	PNode owner= node->getOwner();
	const Nodes & out= *owner->getOut();
	for(Nodes::const_iterator it= out.begin(); it!=out.end(); ++it) {
		int i= FromString<int>(*(*it)->getName());
		if(shift > 0) {
			if(index < i && i <= index+shift)
				setName(*it, ToString(i-1));
		} else {
			if(index+shift <= i && i < index)
				setName(*it, ToString(i+1));
		}
	}
	setName(node, ToString(index+shift));
}

void Model::ensureOut(PNode node, const std::vector<std::string> & names) {
	const Nodes & out= *node->getOut();
	Nodes del;
	for(Nodes::const_iterator it= out.begin(); it!=out.end(); ++it) {
		const std::string & name= *(*it)->getName();
		bool found= std::find(names.begin(), names.end(), name) != names.end();
		if(!found)
			del.push_back(*it);
	}
	for(Nodes::iterator it= del.begin(); it!=del.end(); ++it)
		remove(*it);
}

void Model::clearLink(PNode node) {
	CHECK(node->getRole() == nrLink);
	if(PNode target= node->getLink()) {
		if(gatherLinks)
			linksCleared.push_back(node);
		setModified(node);
		op_setLink(node, PNode());
		op_eraseIn(target, node);
		if(!node->isWeakLink())
			setModified(target);
	}
}

void Model::setLink(PNode node, PNode target) {
	CHECK(target);
	PNode oldTarget= node->getLink();
	if(oldTarget != target) {
		if(oldTarget)
			clearLink(node);
		op_setLink(node, target);
		op_pushIn(target, node);
		setModified(node);
		if(!node->isWeakLink())
			setModified(target);
	}
}

void Model::setScalar(PNode node, Any value) {
	CHECK(node->getRole() == nrScalar);
	CHECK(value && value->getType() == node->getType());
	Any old= node->getValue();
	if(!Equal(value, old)) {
		op_setValue(node, value);
		setModified(node);
	}
}

void Model::setScalar(PNode owner, const std::string & name, Any value) {
	PNode node= find(owner, name);
	CHECK(node);
	setScalar(node, value);
}

void Model::setName(PNode node, const std::string & name) {
	if(*node->getName() != name) {
		op_setName(node, name);
		setModified(node);
		if(node->getOwner()==root) {
			const Nodes & links= *node->getIn();
			for(Nodes::const_iterator it= links.begin(); it!=links.end(); ++it)
				setModified(*it);
		}
	}
}

void Model::setMeta(PNode node, const Glib::ustring & meta) {
	if(*node->getMeta() != meta) {
		op_setMeta(node, meta);
		setModified(node);
	}
}

void Model::setChildNode(PNode widget, PNode child) {
	PNode old= findChildNode(widget);
	if(old) {
		PNode link= find(old, "widget");
		clearLink(link);
		vectorRemove(old);
	}
	PNode link= find(child, "widget");
	setLink(link, widget);
}

void Model::clearState() {
	dirty= false;
	clearModified(root);
	changeType= ctNone;
}

void Model::begin(UpdateMode um) {
	CHECK(mode==umNone);
	CHECK(!dirty);
	CHECK(!isModified());
	CHECK(changeType==ctNone);
	CHECK(log.empty());
	mode= um;
	dirty= true;
	if(mode==umLoad) {
		CHECK(empty());
		CHECK(emptyHistory());
	}
}

void Model::clean() {
	CHECK(mode==umNormal || mode==umClear);
	while(!root->getOut()->empty())
		rootRemove(root->getOut()->front());
}

void Model::seek(int ptr) {
	CHECK(mode==umSeek);
	CHECK(0 <= ptr && ptr <= hsize);
	int count= hpointer - ptr;
	if(count > 0)
		for(int i=0; i<count; ++i)
			undo();
	if(count < 0)
		for(int i=0; i<-count; ++i)
			redo();
}

void Model::rollback() {

	switch(mode) {

		case umNormal:
		case umPaste:
		{
			for(int i=log.size()-1; i>=0; --i)
				undo(log[i]);
			log.clear();
			mode= umNone;
			break;
		}

		case umLoad:
		{
			commit();
			bool wasModified= isModified();
			clearState();
			begin(umClear);
			clean();
			if(wasModified)
				tagModified();
			commit();
			break;
		}

		default:
			CHECK(false);

	}

}

void Model::commit() {

	switch(mode) {

		case umNormal:
		case umPaste:
			if(!redundantLog())
				appendHistory();
			log.clear();
			break;

		case umLoad:
			CHECK(testChange(ctAppend | ctEdit));
			CHECK(log.empty());
			CHECK(emptyHistory());
			break;

		case umClear:
			CHECK(testChange(ctRemove));
			CHECK(log.empty());
			clearHistory();
			break;

		case umSeek:
			CHECK(testChange(ctNone));
			CHECK(log.empty());
			break;

		default:
			CHECK(false);

	}

	mode= umNone;
}

void Model::clear() {
	begin(umClear);
	clean();
	commit();
}

bool Model::getLinksCleared() {
	gatherLinks= false;
	for(Nodes::iterator it= linksCleared.begin(); it!=linksCleared.end(); ++it)
		if(!(*it)->isInactive()) {
			linksCleared.clear();
			return true;
		}
	linksCleared.clear();
	return false;
}

bool Model::isChildrenVector(PNode node) {
	return node && node->getRole()==nrVector && *node->getName()=="children";
}

void Model::getNested(PNode node, Nodes * nodes, bool all) {
	CHECK(node);
	if(node->getRole()!=nrLink && node->getRole()!=nrScalar) {

		const Nodes & out= *node->getOut();
		for(Nodes::const_iterator it= out.begin(); it!=out.end(); ++it) {
			if((*it)->getRole()==nrEntity) {
				nodes->push_back(*it);
				if(all)
					getNested(*it, nodes, true);
			} else
				getNested(*it, nodes, all);

		}

	}
}

void Model::getChildren(PNode node, Nodes * nodes, bool all) {
	PNode children= find(node, "children");
	if(isChildrenVector(children)) {
		const Nodes & out= *children->getOut();
		for(Nodes::const_iterator it= out.begin(); it!=out.end(); ++it) {
			PNode child= getWidgetNode(*it);
			if(child) {
				nodes->push_back(child);
				if(all)
					getChildren(child, nodes, true);
			}
		}
	}
}

void Model::getSlaves(PNode node, Nodes * nodes, bool all) {
	CHECK(node);
	if(node->getRole()==nrLink) {

		if(!node->isWeakLink()) {
			PNode target= node->getLink();
			if(target) {
				nodes->push_back(target);
				if(all)
					getSlaves(target, nodes, true);
			}
		}

	} else if(node->getRole()!=nrScalar) {

		const Nodes & out= *node->getOut();
		for(Nodes::const_iterator it= out.begin(); it!=out.end(); ++it)
			getSlaves(*it, nodes, all);

	}
}

void Model::query(PNode node, TypeId baseType, Nodes * nodes, bool all) {
	if(node->getRole()!=nrScalar && node->getRole()!=nrLink) {
		const Nodes & out= *node->getOut();
		for(Nodes::const_iterator it=out.begin(); it!=out.end(); ++it) {
			PNode node= *it;
			if(
				(node->getRole()==nrScalar || node->getRole()==nrEntity) &&
				Palette::get()->typeIsA(node->getType(), baseType)
			)
				nodes->push_back(node);
			if(all)
				query(node, baseType, nodes, all);
		}
	}
}

PNode Model::createNode(NodeRole role, TypeId type, StateFlags sf) {
	return Node::create(role, type, sf);
}

void Model::op_setOwner(PNode node, PNode owner) {
	CHECK(!readOnly);
	CHECK(mode==umNormal || mode==umPaste || (owner ? mode==umLoad : mode==umClear));
	PNode oldowner= node->getOwner();
	node->setOwner(owner);
	changeType |= owner ? ctAppend : ctRemove;
	if(mode==umNormal || mode==umPaste)
		log.push_back(Operation::create(ocSetOwner, node, oldowner, owner));
}

void Model::op_modified(PNode node) {
	CHECK(!readOnly);
	CHECK(mode!=umNone);
	node->setModified(true);
	if(mode==umNormal || mode==umPaste)
		log.push_back(Operation::create(ocModified, node));
}

void Model::op_setName(PNode node, const std::string & newname) {
	CHECK(!readOnly);
	std::string oldname= *node->getName();
	CHECK(mode==umNormal || ((mode==umLoad || mode==umPaste) && oldname.empty()) || mode==umClear);
	node->setName(newname);
	if(mode==umClear)
		CHECK(node->getOwner()->getRole()==nrVector);
	else
		changeType |= oldname.empty() ? ctAppend : ctEdit;
	if(mode==umNormal || mode==umPaste)
		log.push_back(Operation::create(ocSetName, node, oldname, newname));
}

void Model::op_setMeta(PNode node, const Glib::ustring & newmeta) {
	CHECK(!readOnly);
	Glib::ustring oldmeta= *node->getMeta();
	CHECK(mode==umNormal || ((mode==umLoad || mode==umPaste) && oldmeta.empty()));
	node->setMeta(newmeta);
	changeType |= oldmeta.empty() ? ctAppend : ctEdit;
	if(mode==umNormal || mode==umPaste)
		log.push_back(Operation::create(ocSetMeta, node, oldmeta, newmeta));
}

void Model::op_setValue(PNode node, Any newvalue) {
	CHECK(!readOnly);
	Any oldvalue= node->getValue();
	CHECK(mode==umNormal || mode==umPaste || mode==umLoad || (mode==umSeek && node->isNoUndo()));
	node->setValue(newvalue);
	if(!node->isNoUndo())
		changeType |= !oldvalue ? ctAppend : ctEdit;
	if((mode==umNormal || mode==umPaste) && !node->isNoUndo())
		log.push_back(Operation::create(ocSetValue, node, oldvalue, newvalue));
}

void Model::op_pushIn(PNode node, PNode push) {
	CHECK(!readOnly);
	CHECK(mode==umNormal || mode==umPaste || mode==umLoad);
	node->pushIn(push);
	changeType |= ctAppend;
	if(mode==umNormal || mode==umPaste)
		log.push_back(Operation::create(ocPushIn, node, push));
}

void Model::op_pushOut(PNode node, PNode push) {
	CHECK(!readOnly);
	CHECK(mode==umNormal || mode==umPaste || mode==umLoad);
	node->pushOut(push);
	changeType |= ctAppend;
	if(mode==umNormal || mode==umPaste)
		log.push_back(Operation::create(ocPushOut, node, push));
}

void Model::op_eraseIn(PNode node, PNode erase) {
	CHECK(!readOnly);
	CHECK(mode==umNormal || mode==umPaste || mode==umClear);
	node->eraseIn(erase);
	changeType |= ctRemove;
	if(mode==umNormal || mode==umPaste)
		log.push_back(Operation::create(ocEraseIn, node, erase));
}

void Model::op_eraseOut(PNode node, PNode erase) {
	CHECK(!readOnly);
	CHECK(mode==umNormal || mode==umPaste || mode==umClear);
	node->eraseOut(erase);
	changeType |= ctRemove;
	if(mode==umNormal || mode==umPaste)
		log.push_back(Operation::create(ocEraseOut, node, erase));
}

void Model::op_setLink(PNode node, PNode link) {
	CHECK(!readOnly);
	CHECK(mode==umNormal || mode==umPaste || (link ? mode==umLoad : mode==umClear));
	PNode oldlink= node->getLink();
	node->setLink(link);
	changeType |= link ? ctAppend : ctRemove;
	if(mode==umNormal || mode==umPaste)
		log.push_back(Operation::create(ocSetLink, node, oldlink, link));
}

void Model::setModified(PNode node) {

	PNode N= node;

	do {

		if(N->isModified())
			break;

		op_modified(N);

		N= N->getOwner();

	} while(N);

}

void Model::clearModified(PNode node) {
	if(node->isModified()) {
		node->setModified(false);
		if(node->getRole()!=nrLink && node->getRole()!=nrScalar) {
			const Nodes & out= *node->getOut();
			for(Nodes::const_iterator it= out.begin(); it!=out.end(); ++it)
				clearModified(*it);
		}
	}
}

bool Model::redundantLog() {
	for(int i=0; i<SIGNED(log.size()); ++i)
		if(log[i]->code != ocModified)
			return false;
	return true;
}

void Model::undo() {
	CHECK(hpointer > 0);
	--pointer;
	--hpointer;
	const Log & ops= *pointer;
	for(int i=ops.size()-1; i>=0; --i)
		undo(ops[i]);
}

void Model::redo() {
	CHECK(hpointer < hsize);
	const Log & ops= *pointer;
	for(int i=0; i<SIGNED(ops.size()); ++i)
		redo(ops[i]);
	++pointer;
	++hpointer;
}

void Model::appendHistory() {
	CHECK(hpointer <= hsize);
	for(int i=0; i<hsize-hpointer; ++i)
		history.pop_back();
	history.push_back(log);
	pointer= history.end();
	++hpointer;
	hsize= hpointer;
}

void Model::clearHistory() {
	history.clear();
	pointer= history.end();
	hpointer= 0;
	hsize= 0;
}

void Model::undo(POperation op) {

	switch(op->code) {

		case ocSetOwner: {
			POperationNode opc= POperationNode::cast_static(op);
			CHECK(opc->node->getOwner()==opc->newnode);
			opc->node->setOwner(opc->oldnode);
			if(!opc->node->isInactive())
				opc->node->setModified(true);
			break;
		}

		case ocModified:
			op->node->setModified(true);
			break;

		case ocSetName: {
			POperationName opc= POperationName::cast_static(op);
			CHECK(*opc->node->getName()==opc->newname);
			opc->node->setName(opc->oldname);
			break;
		}

		case ocSetMeta: {
			POperationMeta opc= POperationMeta::cast_static(op);
			CHECK(*opc->node->getMeta()==opc->newmeta);
			opc->node->setMeta(opc->oldmeta);
			break;
		}

		case ocSetValue: {
			POperationValue opc= POperationValue::cast_static(op);
			CHECK(opc->node->getValue()==opc->newvalue);
			opc->node->setValue(opc->oldvalue);
			break;
		}

		case ocPushIn: {
			POperationLink opc= POperationLink::cast_static(op);
			opc->node->eraseIn(opc->link);
			break;
		}

		case ocPushOut: {
			POperationLink opc= POperationLink::cast_static(op);
			opc->node->eraseOut(opc->link);
			break;
		}

		case ocEraseIn: {
			POperationLink opc= POperationLink::cast_static(op);
			opc->node->pushIn(opc->link);
			break;
		}

		case ocEraseOut: {
			POperationLink opc= POperationLink::cast_static(op);
			opc->node->pushOut(opc->link);
			break;
		}

		case ocSetLink: {
			POperationNode opc= POperationNode::cast_static(op);
			CHECK(opc->node->getLink()==opc->newnode);
			opc->node->setLink(opc->oldnode);
			break;
		}

		default:
			CHECK(false);
	}

}

void Model::redo(POperation op) {

	switch(op->code) {

		case ocSetOwner: {
			POperationNode opc= POperationNode::cast_static(op);
			CHECK(opc->node->getOwner()==opc->oldnode);
			opc->node->setOwner(opc->newnode);
			if(!opc->node->isInactive())
				opc->node->setModified(true);
			break;
		}

		case ocModified:
			op->node->setModified(true);
			break;

		case ocSetName: {
			POperationName opc= POperationName::cast_static(op);
			CHECK(*opc->node->getName()==opc->oldname);
			opc->node->setName(opc->newname);
			break;
		}

		case ocSetMeta: {
			POperationMeta opc= POperationMeta::cast_static(op);
			CHECK(*opc->node->getMeta()==opc->oldmeta);
			opc->node->setMeta(opc->newmeta);
			break;
		}

		case ocSetValue: {
			POperationValue opc= POperationValue::cast_static(op);
			CHECK(opc->node->getValue()==opc->oldvalue);
			opc->node->setValue(opc->newvalue);
			break;
		}

		case ocPushIn: {
			POperationLink opc= POperationLink::cast_static(op);
			opc->node->pushIn(opc->link);
			break;
		}

		case ocPushOut: {
			POperationLink opc= POperationLink::cast_static(op);
			opc->node->pushOut(opc->link);
			break;
		}

		case ocEraseIn: {
			POperationLink opc= POperationLink::cast_static(op);
			opc->node->eraseIn(opc->link);
			break;
		}

		case ocEraseOut: {
			POperationLink opc= POperationLink::cast_static(op);
			opc->node->eraseOut(opc->link);
			break;
		}

		case ocSetLink: {
			POperationNode opc= POperationNode::cast_static(op);
			CHECK(opc->node->getLink()==opc->oldnode);
			opc->node->setLink(opc->newnode);
			break;
		}

		default:
			CHECK(false);
	}

}

}
