/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2009 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#include "headers.h"
#include "forward.h"
#include "session.h"

#include "editors.h"
#include "views.h"
#include "palette.h"
#include "document.h"

namespace Crow {

// Session ---------------------------------------------------------------------

PSession Session::create() {
	return NewRefPtr(new Session());
}

SessionRole Session::getRoleForNodes(const Nodes & nodes) {
	Nodes::const_iterator it= nodes.begin();
	SessionRole role= Session::getRoleForNode(*it);
	for(++it; it!=nodes.end(); ++it)
		if(role!=Session::getRoleForNode(*it))
			return srNone;
	return role;
}

SessionRole Session::getRoleForNode(PNode node) {
	switch(node->getRole()) {
		case nrScalar:
			return srScalar;
		case nrVector:
			return srVector;
		case nrLink:
			return srObject;
		case nrEntity:
			return srObject;
		default:
			return srNone;
	}
}

Session::Properties Session::findProperties(Controller * controller, const Nodes & nodes) {
	Properties properties;
	properties.reserve(nodes.size());
	for(Nodes::const_iterator it=nodes.begin(); it!=nodes.end(); ++it) {
		Property * property= controller->findViewProperty(*it);
		properties.push_back(property);
	}
	return properties;
}

EditorId Session::getEditorId(const Properties & properties) {
	EditorId editor= InvalidEditorId;
	for(int i=0; i<SIGNED(properties.size()); ++i) {
		TypeId E= properties[i]->getEditor();
		if(!editor || editor==E)
			editor= E;
		else
			return InvalidEditorId;
	}
	return editor;
}

TypeId Session::getPropertyType(const Properties & properties) {
	TypeId type= properties[0]->getType();
	for(int i=1; i<SIGNED(properties.size()); ++i)
		if(type != properties[i]->getType())
			return InvalidTypeId;
	return type;
}

void Session::init(const Nodes & nodes, const IdPath & path, SessionManager * manager) {
	this->nodes= nodes;
	this->path= path;
	this->manager= manager;
	role= Session::getRoleForNodes(nodes);
}

void Session::done() {
	if(editor) {
		editor->done();
		editor= PEditor();
	}
	path.clear();
	nodes.clear();
	manager= NULL;
	operation= soRemove;
	role= srNone;
}

PEditor Session::getEditor() {
	return editor;
}

void Session::setEditor(PEditor editor) {
	this->editor= editor;
}

bool Session::isMultiple() {
	return nodes.size() > 1;
}

std::string Session::getName() {
	CHECK(!path.empty());
	return path.back().toString();
}

bool Session::modelEditable() {
	Properties properties= getProperties();
	for(int i=0; i<SIGNED(properties.size()); ++i)
		if(!properties[i]->canEdit())
			return false;
	return true;
}

bool Session::canEdit() {
	return modelEditable() && !isFuzzy();
}

bool Session::isModified() {
	for(Nodes::iterator it= nodes.begin(); it!=nodes.end(); ++it)
		if((*it)->isModified())
			return true;
	return false;
}

bool Session::isTouched() {
	for(Nodes::iterator it= nodes.begin(); it!=nodes.end(); ++it)
		if(manager->getController()->isTouched(*it))
			return true;
	return false;
}

bool Session::hasToSave() {
	for(Nodes::iterator it= nodes.begin(); it!=nodes.end(); ++it)
		if(manager->getController()->willSave(*it).first)
			return true;
	return false;
}

bool Session::canDefault() {
	for(Nodes::iterator it= nodes.begin(); it!=nodes.end(); ++it)
		if(manager->getController()->canDefault(*it))
			return true;
	return false;
}

bool Session::isElement() {
	if(isMultiple())
		return false;
	PNode owner= getNode1()->getOwner();
	return owner->getRole()==nrVector;
}

bool Session::hasMasterLink() {
	if(role==srObject)
		for(Nodes::iterator it= nodes.begin(); it!=nodes.end(); ++it) {
			const PNode & N= *it;
			if(N->getRole()==nrLink && N->getLink() && !N->isWeakLink())
				return true;
		}
	return false;
}

bool Session::hasVectorWithMasterLink() {
	if(role==srVector)
		for(Nodes::iterator it= nodes.begin(); it!=nodes.end(); ++it) {
			CHECK((*it)->getRole()==nrVector);
			for(Nodes::const_iterator jt= (*it)->getOut()->begin(); jt!=(*it)->getOut()->end(); ++jt) {
				const PNode & N= *jt;
				if(N->getRole()==nrLink && N->getLink() && !N->isWeakLink())
					return true;
			}
		}
	return false;
}

bool Session::areEmptyLinks() {
	CHECK(role==srObject);
	for(Nodes::iterator it= nodes.begin(); it!=nodes.end(); ++it)
		if((*it)->getRole()!=nrLink || (*it)->getLink())
			return false;
	return true;
}

TypeId Session::getFinalType() {
	CHECK(role==srScalar);
	Properties properties= getProperties();
	TypeId type= Session::getPropertyType(properties);
	CHECK(type);
	return type;
}

TypeId Session::getVectorType() {
	CHECK(role==srVector);
	Properties properties= getProperties();
	TypeId type= Session::getPropertyType(properties);
	CHECK(type);
	return type;
}

Any Session::getScalar() {
	CHECK(role==srScalar);
	Nodes::iterator it= nodes.begin();
	Any value= (*it)->getValue();
	const Glib::ustring & meta= *(*it)->getMeta();
	for(++it; it!=nodes.end(); ++it)
		if(!(*it)->getValue()->equals(value) || *(*it)->getMeta()!=meta)
			return Any();
	return value;
}

Glib::ustring Session::getMeta() {
	CHECK(role==srScalar);
	Nodes::iterator it= nodes.begin();
	Glib::ustring meta= *(*it)->getMeta();
	for(++it; it!=nodes.end(); ++it)
		CHECK(*(*it)->getMeta()==meta);
	return meta;
}

TypeId Session::getObjectBaseType() {
	CHECK(role==srObject);
	Properties properties= getProperties();
	TypeId type= Session::getPropertyType(properties);
	return type;
}

PNode Session::getLinkTarget() {
	CHECK(role==srObject);
	PNode target;
	for(Nodes::iterator it= nodes.begin(); it!=nodes.end(); ++it)
		if((*it)->getRole()==nrLink) {
			PNode T= (*it)->getLink();
			if(T && (!target || target==T))
				target= T;
			else
				return PNode();
		} else
			return PNode();
	return target;
}

TypeId Session::getEntityType() {
	CHECK(role==srObject);
	TypeId type= InvalidTypeId;
	for(Nodes::iterator it= nodes.begin(); it!=nodes.end(); ++it)
		if((*it)->getRole()==nrEntity) {
			TypeId T= (*it)->getType();
			if(!type || type==T)
				type= T;
			else
				return InvalidTypeId;
		} else
			return InvalidTypeId;
	return type;
}

EditorId Session::getEditorId() {
	Properties properties= getProperties();
	if(!isElement())
		return Session::getEditorId(properties);
	else {
		CHECK(properties.size()==1);
		return properties.front()->getElementEditor();
	}
}

bool Session::isFuzzy() {
	switch(role) {

		case srScalar:
			return !getScalar();

		case srVector: {
			Properties properties= getProperties();
			return !Session::getPropertyType(properties);
		}

		case srObject:
			return !(getLinkTarget() || getEntityType() || areEmptyLinks());

		default:
			CHECK(role==srNone);
			CHECK(false);
			return true;
	}
}

void Session::resetScalar() {
	CHECK(role==srScalar);
	Properties properties= getProperties();
	int i=0;
	for(Nodes::iterator it= nodes.begin(); it!=nodes.end(); ++it, ++i)
		if(manager->getController()->canDefault(*it)) {
			Any dvalue= properties[i]->getDefault();
			manager->getModel()->setScalar(*it, dvalue);
			manager->getModel()->setMeta(*it, "");
		}
}

void Session::resetVector() {
	CHECK(role==srVector);
	for(Nodes::iterator it= nodes.begin(); it!=nodes.end(); ++it)
		manager->getModel()->vectorClear(*it);
}

void Session::resetObject() {
	Properties properties= getProperties();
	CHECK(role==srObject);
	Model * model= manager->getModel();
	int i=0;
	for(Nodes::iterator it= nodes.begin(); it!=nodes.end(); ++it, ++i)
		if(manager->getController()->canDefault(*it)) {
			if((*it)->getRole()!=nrLink) {
				PNode owner= (*it)->getOwner();
				model->remove(*it);
				model->createLink(owner, *(*it)->getName(), properties[i]->getNodeState());
			} else
				model->clearLink(*it);
		}
}

void Session::setScalarOnly(Any value) {
	CHECK(role==srScalar && modelEditable());
	for(Nodes::iterator it= nodes.begin(); it!=nodes.end(); ++it)
		manager->getModel()->setScalar(*it, value);
}

void Session::setScalarAndMeta(Any value, const Glib::ustring & meta) {
	CHECK(role==srScalar && modelEditable());
	for(Nodes::iterator it= nodes.begin(); it!=nodes.end(); ++it) {
		manager->getModel()->setScalar(*it, value);
		manager->getModel()->setMeta(*it, meta);
	}
}

void Session::setObjectAsLink(PNode object) {
	Properties properties= getProperties();
	CHECK(role==srObject && modelEditable());
	Model * model= manager->getModel();
	int i=0;
	for(Nodes::iterator it= nodes.begin(); it!=nodes.end(); ++it, ++i) {
		PNode node= *it;
		if(node->getRole()!=nrLink) {
			PNode owner= node->getOwner();
			model->remove(node);
			node= model->createLink(owner, *node->getName(), properties[i]->getNodeState());
		}
		model->setLink(node, object);
	}
}

void Session::setObjectAsEntity(TypeId type) {
	CHECK(role==srObject && modelEditable());
	Model * model= manager->getModel();
	for(Nodes::iterator it= nodes.begin(); it!=nodes.end(); ++it) {
		PNode owner= (*it)->getOwner();
		if((*it)->getRole()==nrLink)
			model->clearLink(*it);
		model->remove(*it);
		model->createEntity(owner, type, *(*it)->getName());
	}
}

Property * Session::getProperty1() {
	CHECK(!isMultiple());
	Properties properties= getProperties();
	return properties[0];
}

PNode Session::getNode1() {
	CHECK(!isMultiple());
	return nodes.front();
}

Session::Session() {
	role= srNone;
	manager= NULL;
	operation= soRemove;
}

bool Session::hasFlags(StateFlags sf) {
	Properties properties= getProperties();
	for(int i=0; i<SIGNED(properties.size()); ++i)
		if(properties[i]->hasFlags(sf))
			return true;
	return false;
}

bool Session::haveFlags(StateFlags sf) {
	Properties properties= getProperties();
	for(int i=0; i<SIGNED(properties.size()); ++i)
		if(!properties[i]->hasFlags(sf))
			return false;
	return true;
}

Session::Properties Session::getProperties() {
	return Session::findProperties(manager->getController(), nodes);
}

// SessionSupplier -------------------------------------------------------------

SessionSupplier::SessionSupplier() {
	manager= NULL;
}

void SessionSupplier::updateSessions() {

	CHECK(found.empty());
	findNodes();

	for(Sessions::iterator it= sessions.begin(); it!=sessions.end(); ++it)
		it->second->operation= soRemove;

	for(NodeMap::iterator it= found.begin(); it!=found.end(); ++it) {

		const IdPath & path= it->first;
		const Nodes & nodes= it->second;
		Sessions::iterator sit= sessions.find(path);

		if(sit!=sessions.end()) {
			PSession session= sit->second;
			if(!ListsEqualSorted(*session->getNodes(), nodes)) {

				sessions.erase(sit);
				stopSession(session);

			} else {
				session->operation= soUpdate;
				continue;
			}

		}

		PSession session= createSession(nodes, path);
		sessions.insert(std::make_pair(path, session));
		session->operation= soInsert;

	}

	for(Sessions::iterator it= sessions.begin(); it!=sessions.end(); )
		if(it->second->operation==soRemove) {
			stopSession(it->second);
			it= EraseAndNext(sessions, it);
		} else
			++it;

	for(Sessions::iterator it= sessions.begin(); it!=sessions.end(); ++it) {
		PSession session= it->second;
		if(session->operation==soInsert)
			session->getEditor()->created();
	}

	for(Sessions::iterator it= sessions.begin(); it!=sessions.end(); ++it) {
		PSession session= it->second;
		CHECK(session->operation==soInsert || session->operation==soUpdate);
		sessionEvent(session);
	}

	found.clear();
}

void SessionSupplier::removeSessions() {
	for(Sessions::iterator it= sessions.begin(); it!=sessions.end(); ++it) {
		it->second->operation= soRemove;
		stopSession(it->second);
	}
	sessions.clear();
}

void SessionSupplier::editorsLoad() {
	for(Sessions::iterator it= sessions.begin(); it!=sessions.end(); ++it) {
		PSession session= it->second;
		if(session->operation==soInsert) {
			session->getEditor()->load(false);
			afterLoad(session);
		} else {
			CHECK(session->operation==soUpdate);
			if(session->isModified() || session->isTouched()) {
				session->getEditor()->load(true);
				afterLoad(session);
			}
		}
	}
}

void SessionSupplier::applySettings() {
	for(Sessions::iterator it= sessions.begin(); it!=sessions.end(); ++it)
		it->second->getEditor()->applySettings();
}

void SessionSupplier::storeSettings() {
	for(Sessions::iterator it= sessions.begin(); it!=sessions.end(); ++it)
		it->second->getEditor()->storeSettings();
}

PSession SessionSupplier::findSession(const IdPath & path) {
	for(Sessions::iterator it= sessions.begin(); it!=sessions.end(); ++it) {
		PSession session= it->second;
		if(*session->getPath()==path)
			return session;
	}
	return PSession();
}

PSession SessionSupplier::findSession(PNode node) {
	for(Sessions::iterator it= sessions.begin(); it!=sessions.end(); ++it) {
		PSession session= it->second;
		if(!session->isMultiple() && session->getNode1()==node)
			return session;
	}
	return PSession();
}

void SessionSupplier::append(PNode cut, PNode node) {
	IdPath path= manager->getModel()->getNodePath(node, cut);
	append(path, node);
}

void SessionSupplier::append(const IdPath & path, PNode node) {
	Nodes & nodes= found[path];
	nodes.push_back(node);
}

PSession SessionSupplier::createSession(const Nodes & nodes, const IdPath & path) {
	PSession session= Session::create();

	session->init(nodes, path, manager);
	session->setEditor(createEditor(session));

	session->editor->init(session);

	return session;
}

PEditor SessionSupplier::createEditor(PSession session) {
	return PEditor();
}

void SessionSupplier::stopSession(PSession session) {
	session->getEditor()->storeSettings();
	session->getEditor()->deleted();
	sessionEvent(session);
	session->done();
}

// HierarchySessionSupplier ----------------------------------------------------

void HierarchySessionSupplier::findNodes() {
	append(IdPath(), getManager()->getModel()->getRoot());
}

PEditor HierarchySessionSupplier::createEditor(PSession session) {
	PEditor editor= NewRefPtr<HierarchyEditor>();
	editor->connect_popup_menu(sigc::mem_fun(*this, &HierarchySessionSupplier::popupMenu));
	return editor;
}

// CanvasSessionSupplier -------------------------------------------------------

void CanvasSessionSupplier::findNodes() {
	PNode node= getManager()->raiseSelectionParent();
	if(node || !getManager()->getSelection()->empty())
		current= node;
	if(current && !current->isInactive()) {
		PNode master= getManager()->getModel()->raiseMaster(current);
		append(IdPath(1, *master->getName()), current);
	}
}

PEditor CanvasSessionSupplier::createEditor(PSession session) {
	EntityEntry * entry= GetEntry<EntityEntry>(session->getEntityType());
	PEditor editor= Palette::get()->createEditor(entry->getCanvasEditor());
	PCanvasEditor ceditor= PCanvasEditor::cast_dynamic(editor);
	ceditor->connect_pasting_finished(sigc::mem_fun(*this, &CanvasSessionSupplier::pastingFinished));
	ceditor->connect_status_changed(sigc::mem_fun(*this, &CanvasSessionSupplier::statusChanged));
	ceditor->connect_popup_menu(sigc::mem_fun(*this, &CanvasSessionSupplier::popupMenu));
	return ceditor;
}

PCanvasEditor CanvasSessionSupplier::getEditor() {
	if(!getSessions()->empty()) {
		PEditor editor= getSessions()->begin()->second->getEditor();
		return PCanvasEditor::cast_dynamic(editor);
	} else
		return PCanvasEditor();
}

bool CanvasSessionSupplier::startPasting(const Glib::ustring & data, bool normal) {
	PCanvasEditor editor= getEditor();
	if(editor)
		return editor->startPasting(data, normal);
	else
		return false;
}

void CanvasSessionSupplier::stopPasting() {
	PCanvasEditor editor= getEditor();
	if(editor)
		editor->stopPasting();
}

// SessionSupplierPropertiesBase -----------------------------------------------

void SessionSupplierPropertiesBase::findProperties(PNode cut, PNode node) {

	bool showHidden= getManager()->getController()->getContext()->getKeyValue<bool>("show_hp");

	const Nodes & out= *node->getOut();
	for(Nodes::const_iterator it= out.begin(); it!=out.end(); ++it) {

		bool skipSignals= false;
		if(*(*it)->getName()=="signals") {
			PNode object= (*it)->getOwner();
			PNode owner= object->getOwner();
			skipSignals= owner == getManager()->getModel()->getRoot();
		}

		bool skipHidden= false;
		if(getManager()->getController()->findViewProperty(*it)->isHidden())
			skipHidden= !showHidden;

		if(!skipSignals && !skipHidden) {

			append(cut, *it);

			NodeRole nrole= (*it)->getRole();

			if(nrole==nrVector && getManager()->getSelection()->size()==1) {
				const Nodes & vec= *(*it)->getOut();
				for(Nodes::const_iterator vit= vec.begin(); vit!=vec.end(); ++vit) {
					append(cut, *vit);
					if((*vit)->getRole()==nrEntity)
						findProperties(cut, *vit);
				}
			}

			if(nrole==nrEntity)
				findProperties(cut, *it);

		}

	}

}

void SessionSupplierPropertiesBase::screenNodes() {
	int count= getManager()->getSelection()->size();
	if(count > 1) {
		for(NodeMap::iterator it= getFound()->begin(); it!=getFound()->end(); ) {
			const Nodes & nodes= it->second;
			if(SIGNED(nodes.size())!=count || !mergeable(nodes))
				it= EraseAndNext(*getFound(), it);
			else
				++it;
		}
	}
}

bool SessionSupplierPropertiesBase::mergeable(const Nodes & nodes) {
	SessionRole role= Session::getRoleForNodes(nodes);
	switch(role) {

		case srNone:
			return false;

		case srVector:
		case srScalar:
		case srObject:
		{
			Session::Properties properties= Session::findProperties(getManager()->getController(), nodes);
			return
				Session::getEditorId(properties) &&
				Session::getPropertyType(properties);
		}

		default:
			CHECK(false);
			return false;
	}
}

PEditor SessionSupplierPropertiesBase::createPropertyEditor(PSession session) {
	return Palette::get()->createEditor(session->getEditorId());
}

// PropertiesSessionSupplier ---------------------------------------------------

void PropertiesSessionSupplier::findNodes() {
	const Nodes & nodes= *getManager()->getSelection();
	for(Nodes::const_iterator it= nodes.begin(); it!=nodes.end(); ++it)
		findProperties(*it, *it);
	screenNodes();
}

PEditor PropertiesSessionSupplier::createEditor(PSession session) {
	return createPropertyEditor(session);
}

// SignalsSessionSupplier ------------------------------------------------------

void SignalsSessionSupplier::findNodes() {
	PNode node= getManager()->getSingleSelection();
	if(node) {
		PNode nodeSignals= getManager()->getModel()->find(node, "signals");
		if(nodeSignals) {
			const Nodes & out= *nodeSignals->getOut();
			for(Nodes::const_iterator it= out.begin(); it!=out.end(); ++it)
				append(nodeSignals, *it);
		}
	}
}

PEditor SignalsSessionSupplier::createEditor(PSession session) {
	return NewRefPtr<EmitterPropertyEditor>();
}

// PackingSessionSupplier ------------------------------------------------------

void PackingSessionSupplier::findNodes() {
	const Nodes & nodes= *getManager()->getSelection();
	for(Nodes::const_iterator it= nodes.begin(); it!=nodes.end(); ++it) {
		PNode nodeChild= getManager()->getModel()->findChildNode(*it);
		if(nodeChild)
			findProperties(nodeChild, nodeChild);
	}
	screenNodes();
}

PEditor PackingSessionSupplier::createEditor(PSession session) {
	return createPropertyEditor(session);
}

// SessionManager --------------------------------------------------------------

SessionManager::SessionManager() :
	states(10)
{
	performing= false;
}

SessionManager::~SessionManager() {

	beginAction();
	signalBeforeSessionsUpdate.emit();
	for(Suppliers::iterator it=suppliers.begin(); it!=suppliers.end(); ++it)
		(*it)->removeSessions();
	signalAfterSessionsUpdate.emit();
	endAction();

	suppliers.clear();

}

void SessionManager::addSupplier(SessionSupplier * supplier) {
	supplier->setManager(this);
	suppliers.push_back(supplier);
}

void SessionManager::acceptSuppliers() {
	beginAction();
	updateSelection(selection, true);
	endAction();
}

bool SessionManager::isSelected(PNode node) {
	return std::find(selection.begin(), selection.end(), node) != selection.end();
}

PNode SessionManager::getSingleSelection() {
	return selection.size()==1 ? selection.front() : PNode();
}

PNode SessionManager::raiseSelectionParent() {
	PNode rsel;
	if(!selection.empty()) {
		Nodes::iterator it= selection.begin();
		rsel= getModel()->raiseParent(*it);
		for(++it; it!=selection.end(); ++it)
			if(rsel != getModel()->raiseParent(*it))
				return PNode();
	}
	return rsel;
}

PNode SessionManager::raiseSelectionMaster() {
	PNode rsel;
	if(!selection.empty()) {
		Nodes::iterator it= selection.begin();
		rsel= getModel()->raiseMaster(*it);
		for(++it; it!=selection.end(); ++it)
			if(rsel != getModel()->raiseMaster(*it))
				return PNode();
	}
	return rsel;
}

void SessionManager::submitSelection(const Nodes & nodes) {
	beginAction();
	updateSelection(nodes);
	endAction();
}

void SessionManager::deleteSelection() {
	begin();
	for(Nodes::const_iterator it=selection.begin(); it!=selection.end(); ++it)
		if(!(*it)->isInactive())
			getModel()->rootRemove(*it);
	commit();
}

void SessionManager::begin(UpdateMode um) {
	beginAction();
	controller.begin(um);
}

void SessionManager::rollback() {
	controller.rollback();
	endAction();
}

void SessionManager::commit() {
	commit(selection);
}

void SessionManager::commit(const Nodes & sel) {
	controller.commit();
	updateSelection(sel, true);
	endAction();
}

void SessionManager::undo() {
	beginAction();
	controller.undo();
	updateSelection(selection, true);
	endAction();
}

void SessionManager::redo() {
	beginAction();
	controller.redo();
	updateSelection(selection, true);
	endAction();
}

bool SessionManager::canUndo() {
	return controller.canUndo();
}

bool SessionManager::canRedo() {
	return controller.canRedo();
}

LoadResult SessionManager::load(const Glib::ustring & data) {
	beginAction();
	LoadResult result;

	try {
		result= controller.load(data);
	} catch(const Glib::Exception &) {
		updateSelection(Nodes(), true);
		endAction();
		throw;
	}

	updateSelection(Nodes(), true);
	endAction();
	return result;
}

void SessionManager::clear() {
	beginAction();
	controller.clear();
	updateSelection(Nodes(), true);
	endAction();
}

void SessionManager::pasteToRoot(const Glib::ustring & data, bool normal) {
	Document document;
	Nodes loaded;
	begin(normal ? umNormal : umPaste);
	bool success= document.load(getModel(), data, loaded);
	if(success)
		commit(getModel()->selectMasters(loaded));
	else {
		rollback();
		throw RuntimeError(*document.getError());
	}
}

Glib::ustring SessionManager::save() {
	return controller.save(*getModel()->getRoot()->getOut());
}

Glib::ustring SessionManager::copy() {
	return controller.save(*getSelection());
}

PSession SessionManager::findSession(PNode node) {
	for(Suppliers::iterator it=suppliers.begin(); it!=suppliers.end(); ++it) {
		PSession session= (*it)->findSession(node);
		if(session)
			return session;
	}
	return PSession();
}

Container * SessionManager::getContainer(PNode node) {
	PGtkContainerView view= controller.getView<GtkContainerView>(node);
	return view ? view->getContainer() : NULL;
}

bool SessionManager::isXYChild(PNode node) {
	PNode parent= getModel()->findParent(node);
	return parent ? dynamic_cast<XYContainer *>(getContainer(parent)) : false;
}

bool SessionManager::isIndexChild(PNode node) {
	PNode parent= getModel()->findParent(node);
	return parent ? dynamic_cast<IndexContainer *>(getContainer(parent)) : false;
}

void SessionManager::applySettings() {
	beginAction();
	updateSessions();
	endAction();
	for(Suppliers::iterator it=suppliers.begin(); it!=suppliers.end(); ++it)
		(*it)->applySettings();
}

void SessionManager::storeSettings() {
	for(Suppliers::iterator it=suppliers.begin(); it!=suppliers.end(); ++it)
		(*it)->storeSettings();
}

void SessionManager::beginAction() {
	signalActionStarted.emit();
	CHECK(!performing);
	performing= true;
}

void SessionManager::endAction() {
	CHECK(performing);
	performing= false;
	getController()->clearState();
	signalActionCompleted.emit();
}

void SessionManager::updateSessions() {
	signalBeforeSessionsUpdate.emit();
	for(Suppliers::iterator it=suppliers.begin(); it!=suppliers.end(); ++it)
		(*it)->updateSessions();
	signalAfterSessionsUpdate.emit();
	for(Suppliers::iterator it=suppliers.begin(); it!=suppliers.end(); ++it)
		(*it)->editorsLoad();
}

bool SessionManager::acceptSelection(const Nodes & nodes) {
	Nodes nsel= nodes;

	for(Nodes::iterator it= nsel.begin(); it!=nsel.end(); )
		if((*it)->isInactive())
			it= nsel.erase(it);
		else
			++it;

	bool equal= ListsEqualSorted(nsel, selection);
	if(!equal) {

		for(Nodes::iterator it= selection.begin(); it!=selection.end(); )
			if(std::find(nsel.begin(), nsel.end(), *it) == nsel.end())
				it= selection.erase(it);
			else
				++it;

		for(Nodes::iterator it= nsel.begin(); it!=nsel.end(); ++it)
			if(std::find(selection.begin(), selection.end(), *it) == selection.end())
				selection.push_back(*it);

		reducedSelection= getModel()->selectParents(selection);
	}

	return !equal;
}

void SessionManager::updateSelection(const Nodes & sel, bool force) {
	bool changed= acceptSelection(sel);
	if(changed || force)
		updateSessions();
	if(changed)
		signalSelectionChanged.emit();
}

}
