/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2009 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#include "headers.h"
#include "forward.h"
#include "value.h"

#include "palette.h"

namespace Crow {

// CAny ------------------------------------------------------------------------

Any CAny::createInt(int data) {
	return CAny::create<int>("int", data);
}

Any CAny::createUnsigned(unsigned data) {
	return CAny::create<unsigned>("unsigned", data);
}

Any CAny::createBool(bool data) {
	return CAny::create<bool>("bool", data);
}

Any CAny::createFloat(float data) {
	return CAny::create<float>("float", data);
}

Any CAny::createDouble(double data) {
	return CAny::create<double>("double", data);
}

Any CAny::createString(const Glib::ustring & data) {
	return CAny::create<Glib::ustring>("string", data);
}

Any CAny::createEnum(TypeId type, int data) {
	CEnum * tany= new CEnum;
	tany->type= type;
	tany->value= data;
	return NewRefPtr(tany);
}

Any CAny::createEnum(const std::string & tname, int data) {
	return createEnum(FindType(tname), data);
}

Any CAny::createFlags(TypeId type, unsigned data) {
	CFlags * tany= new CFlags;
	tany->type= type;
	tany->value= data;
	return NewRefPtr(tany);
}

Any CAny::createFlags(const std::string & tname, unsigned data) {
	return createFlags(FindType(tname), data);
}

Any CAny::createObject(TypeId type, PGlibObject value) {
	if(value) {
		TypeId found= Palette::get()->findType(value);
		CHECK(found == type && type);
	}
	TAnyImpl<PGlibObject> * tany= new TAnyImpl<PGlibObject>;
	tany->type= value ? type : InvalidTypeId;
	tany->value= value;
	return NewRefPtr(tany);
}

Any CAny::createObject(PGlibObject value) {
	TypeId type= Palette::get()->findType(value);
	return createObject(type, value);
}

Any CAny::createObject() {
	TAnyImpl<PGlibObject> * tany= new TAnyImpl<PGlibObject>;
	tany->type= InvalidTypeId;
	tany->value= PGlibObject();
	return NewRefPtr(tany);
}

Any CAny::createVector(const AnyVector & value) {
	CVector * tany= new CVector;
	tany->type= InvalidTypeId;
	tany->value= value;
	return NewRefPtr(tany);
}

Any CAny::createPoint(const Point & data) {
	return CAny::create<Point>("CrowPoint", data);
}

Any CAny::createEmitter(const Emitter & data) {
	return CAny::create<Emitter>("CrowEmitter", data);
}

Any CAny::createFromGValue(TypeId type, const Glib::ValueBase & data) {
	ScalarEntry * sentry= Palette::get()->getEntry<ScalarEntry>(type);
	CHECK(sentry);
	Any value= sentry->createFromGValue(data);
	return value;
}

Any CAny::createFromString(TypeId type, const Glib::ustring & data) {
	ScalarEntry * sentry= Palette::get()->getEntry<ScalarEntry>(type);
	CHECK(sentry);
	Any value= sentry->createFromString(data);
	return value;
}

int CAny::getInt() const {
	return *get<int>();
}

unsigned CAny::getUnsigned() const {
	return *get<unsigned>();
}

bool CAny::getBool() const {
	return *get<bool>();
}

float CAny::getFloat() const {
	return *get<float>();
}

double CAny::getDouble() const {
	return *get<double>();
}

const Glib::ustring * CAny::getString() const {
	return get<Glib::ustring>();
}

int CAny::getEnum() const {
	return *get<int>();
}

unsigned CAny::getFlags() const {
	return *get<unsigned>();
}

PGlibObject CAny::getObject() const {
	return *get<PGlibObject>();
}

const AnyVector * CAny::getVector() const {
	return get<AnyVector>();
}

const Point * CAny::getPoint() const {
	return get<Point>();
}

const Emitter * CAny::getEmitter() const {
	return get<Emitter>();
}

CAny::CAny() {
	type= InvalidTypeId;
}

// CVector ---------------------------------------------------------------------

bool CVector::equals(PObject v) const {
	CVector * pv= dynamic_cast<CVector *>(GetPtr(v));
	if(this==pv)
		return true;
	if(!pv)
		return false;
	if(getValue()->size()!=pv->getValue()->size())
		return false;
	for(int i=0; i<SIGNED(getValue()->size()); ++i)
		if(!(*getValue())[i]->equals((*pv->getValue())[i]))
			return false;
	return true;
}

// CEnum -----------------------------------------------------------------------

Glib::ustring CEnum::toString() const {
	return *Palette::get()->getEntry<EnumEntry>(getType())->getByValue(*getValue());
}

// CFlags ----------------------------------------------------------------------

Glib::ustring CFlags::toString() const {
	std::ostringstream oss;
	oss << std::hex << std::uppercase;
	oss << "0x" << *getValue();
	return oss.str();
}

}
