/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2009 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#ifndef CROW_VALUE_H
#define CROW_VALUE_H

#include "util.h"

namespace Crow {

class CAny : public Object {

public:
	static Any createInt(int data);
	static Any createUnsigned(unsigned data);
	static Any createBool(bool data);
	static Any createFloat(float data);
	static Any createDouble(double data);
	static Any createString(const Glib::ustring & data);
	static Any createEnum(TypeId type, int data);
	static Any createEnum(const std::string & tname, int data);
	static Any createFlags(TypeId type, unsigned data);
	static Any createFlags(const std::string & tname, unsigned data);
	static Any createObject(TypeId type, PGlibObject data);
	static Any createObject(PGlibObject data);
	static Any createObject();
	static Any createVector(const AnyVector & data);
	static Any createPoint(const Point & data);
	static Any createEmitter(const Emitter & emitter);

	template <class T> static Any create(TypeId type, const T & data);
	template <class T> static Any create(const std::string & type, const T & data);

	static Any createFromGValue(TypeId type, const Glib::ValueBase & data);
	static Any createFromString(TypeId type, const Glib::ustring & data);

	TypeId getType() const { return type; }

	int getInt() const;
	unsigned getUnsigned() const;
	bool getBool() const;
	float getFloat() const;
	double getDouble() const;
	const Glib::ustring * getString() const;
	int getEnum() const;
	unsigned getFlags() const;
	PGlibObject getObject() const;
	const AnyVector * getVector() const;
	const Point * getPoint() const;
	const Emitter * getEmitter() const;

	template <class T> Glib::RefPtr<T> getObject() const;
	template <class T> const T * get() const;

	virtual Glib::ustring toString() const = 0;
	virtual Glib::ValueBase toGValue() const = 0;

protected:
	CAny();

	template <class A, class B> static bool equal(A * a, B * b);

private:
	TypeId type;

};

template <class T> class TAnyBase : public CAny {

friend class CAny;

public:
	bool equals(PObject v) const { return CAny::equal(this, GetPtr(Any::cast_dynamic(v))); }
	Glib::ustring toString() const { return ToString<T>(value); }
	Glib::ValueBase toGValue() const { return MakeGValue<T>(value); }

protected:
	const T * getValue() const { return &value; }

private:
	T value;

};

template <class T> class TAnyImpl;

template <class A, class B> bool CAny::equal(A * a, B * b) {
	if(a==b)
		return true;
	if(!a || !b)
		return false;
	if(a->getType() != b->getType())
		return false;
	A * ab= dynamic_cast<A *>(b);
	if(!ab)
		return false;
	return *a->getValue() == *ab->getValue();
}

template <class T> Any CAny::create(TypeId type, const T & data) {
	TAnyImpl<T> * tany= new TAnyImpl<T>;
	tany->type= type;
	tany->value= data;
	return NewRefPtr(tany);
}

template <class T> Any CAny::create(const std::string & type, const T & data) {
	TypeId tid= FindType(type);
	return CAny::create<T>(tid, data);
}

template <class T> const T * CAny::get() const {
	const TAnyImpl<T> * tany= dynamic_cast<const TAnyImpl<T> *>(this);
	CHECK(tany);
	return tany->getValue();
}

template <class T> Glib::RefPtr<T> CAny::getObject() const {
	PGlibObject object= *get<PGlibObject>();
	return Glib::RefPtr<T>::cast_dynamic(object);
}

template <> inline Glib::ustring ToString<AnyVector>(const AnyVector & value) { CHECK(false); return ""; }

template <> class TAnyImpl<int> : public TAnyBase<int> { };
template <> class TAnyImpl<unsigned> : public TAnyBase<unsigned> { };
template <> class TAnyImpl<bool> : public TAnyBase<bool> { };
template <> class TAnyImpl<float> : public TAnyBase<float> { };
template <> class TAnyImpl<double> : public TAnyBase<double> { };
template <> class TAnyImpl<PGlibObject> : public TAnyBase<PGlibObject> { };
template <> class TAnyImpl<AnyVector> : public TAnyBase<AnyVector> { };
template <> class TAnyImpl<Point> : public TAnyBase<Point> { };
template <> class TAnyImpl<Emitter> : public TAnyBase<Emitter> { };
template <> class TAnyImpl<Glib::ustring> : public TAnyBase<Glib::ustring> { };
template <> class TAnyImpl<UniChar> : public TAnyBase<UniChar> { };
template <> class TAnyImpl<Border> : public TAnyBase<Border> { };
template <> class TAnyImpl<Gtk::StockID> : public TAnyBase<Gtk::StockID> { };
template <> class TAnyImpl<Gdk::Color> : public TAnyBase<Gdk::Color> { };

class CVector : public TAnyImpl<AnyVector> {

friend class CAny;

public:
	bool equals(PObject v) const;

};

class CEnum : public TAnyImpl<int> {

friend class CAny;

public:
	Glib::ustring toString() const;

};

class CFlags : public TAnyImpl<unsigned> {

friend class CAny;

public:
	Glib::ustring toString() const;

};

}

#endif
