/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2009 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#ifndef CROW_WIDGETS_H
#define CROW_WIDGETS_H

#include "polytree.h"

namespace Crow {

class HierarchyWidget : public Gtk::Frame {

public:
	HierarchyWidget();

	HierarchyEditorWidget * getEditorWidget();
	void setEditorWidget(EditorWidget *);
	void unsetEditorWidget();

	int getSplit();
	void setSplit(int split);

};

enum PaletteEvent { peRoot, peCanvas, peCancel };

class PaletteWidget : public Gtk::Frame {

public:
	PaletteWidget();

	void setShowDW(bool show_dw);
	void unpush();

	typedef sigc::signal<void, TypeId, PaletteEvent> SignalPaletteEvent;
	sigc::connection connect_palette_event(SignalPaletteEvent::slot_type slot) { return signalPaletteEvent.connect(slot); }

protected:
	void create();
	Gtk::Widget * createPaletteButton(EntityEntry * entry);
	void onEntryClicked(TypeId type, Gtk::ToggleButton * tbutton);

private:
	Gtk::ScrolledWindow scroll;
	Gtk::VBox vBox;
	SignalPaletteEvent signalPaletteEvent;
	Gtk::ToggleButton * pushed;
	Gtk::Widget * debugTab;

};

class CanvasWidget : public Gtk::Frame {

public:
	CanvasWidget();

	void setEditorWidget(EditorWidget *);
	void unsetEditorWidget();

};

class PropertyTree : public Polytree {

public:
	PropertyTree();

	int getSplit();
	void setSplit(int split);

};

class ExplorerWidget : public Gtk::Frame {

public:
	ExplorerWidget();

	enum ExplorerPage { epProperties, epSignals, epPacking };

	PropertyTree * getTreeView(ExplorerPage page);

	int getSplit();
	void setSplit(int split);

	PSession getSelectedRow();
	void setSelectedRow(PSession session);

	ExplorerPage getCurrentPage();
	PropertyTree * getCurrentTree();

	typedef sigc::signal<void> SignalSelectionChanged;
	sigc::connection connect_selection_changed(SignalSelectionChanged::slot_type slot) { return signalSelectionChanged.connect(slot); }

protected:
	void addPage(const std::string & tab);

	void onSelectionChanged(ExplorerPage page);
	void onSwitchPage(GtkNotebookPage *, guint);

private:
	Gtk::Notebook notebook;
	SignalSelectionChanged signalSelectionChanged;
	int prev_page;

};

class SignalSelectDialog : public Gtk::Dialog {

public:
	SignalSelectDialog();

	void setChoices(const SignalsQuery & signals);
	std::string getSelection() { return selection; }

protected:
	void addTreeView(GType gtype, const std::vector<std::string> & names);
	void onSelectionChanged(Gtk::TreeView * treeView);

private:
	bool lock;
	std::string selection;

	Gtk::ScrolledWindow scroll;
	Gtk::VBox vBox;

	Gtk::Button * cancelButton;
	Gtk::Button * okButton;

};

class TextInputDialog : public Gtk::Dialog {

public:
	TextInputDialog();

	Glib::ustring getText();
	void setText(const Glib::ustring & text);

	Translatable getMeta();
	void setMeta(const Translatable & trans);

private:
	Gtk::VPaned paned;
	Gtk::ScrolledWindow scrollText;
	Gtk::TextView text;
	Gtk::Frame frame;
	Gtk::CheckButton check;
	Gtk::Alignment align;
	Gtk::VBox box1;
	Gtk::HBox box2;
	Gtk::VBox box3;
	Gtk::Label label1;
	Gtk::Entry entry;
	Gtk::Label label2;
	Gtk::ScrolledWindow scrollComments;
	Gtk::TextView comments;

};

class IconSelectDialog : public Gtk::Dialog {

public:
	IconSelectDialog();

	void setChoices(const std::vector<std::pair<PGdkPixbuf, Glib::ustring> > & choices);

	Glib::ustring getChoice();
	void setChoice(const Glib::ustring & name);

protected:
	void onSelectionChanged();

private:
	Gtk::ScrolledWindow scroll;
	Gtk::TreeView treeView;

	Gtk::Button * cancelButton;
	Gtk::Button * okButton;

};

class FlagsDialog : public Gtk::Dialog {

public:
	FlagsDialog();

	typedef std::list<std::pair<std::string, Bits> > Values;

	void setOptions(const Values & options);
	Bits getValue();
	void setValue(Bits value);

private:
	Gtk::ScrolledWindow scroll;
	Gtk::TreeView treeView;

};

class TranslatablesDialog : public Gtk::Dialog {

public:
	TranslatablesDialog();

	struct String {
		PNode node;
		IdPath path;
		Glib::ustring string;
		Glib::ustring defval;
		Translatable meta;
		bool operator < (const String & s) const { return path < s.path; }
	};

	typedef std::vector<String> Strings;

	void setStrings(const Strings & strings);
	Strings getStrings();

protected:
	void setupColumn(int column, int width);
	void onInvert();

private:
	Gtk::Button invertButton;
	Gtk::ScrolledWindow scroll;
	Gtk::TreeView treeView;

	Strings set;

};

class DesignWindow : public Gtk::Alignment {

public:
	DesignWindow();

	Glib::PropertyProxy<Glib::ustring> property_title() { return propertyTitle.get_proxy(); }
	Glib::PropertyProxy<Glib::ustring> property_icon_name() { return propertyIconName.get_proxy(); }

protected:
	bool on_expose_event(GdkEventExpose * e);
	void on_size_request(Gtk::Requisition * requisition);

	PPangoLayout createLayout(const Glib::ustring & text, int width);
	Rectangle getCaptionRect(const Rectangle & A);
	Rectangle getButtonRect(const Rectangle & A, int pos);

	void drawIcon(PGdkGC gc, const Rectangle & R);
	void drawMinimize(PGdkGC gc, const Rectangle & R);
	void drawMaximize(PGdkGC gc, const Rectangle & R);
	void drawClose(PGdkGC gc, const Rectangle & R);

	void onTitleChanged();
	void onIconNameChanged();
	void loadIcon();

private:
	int CS, IS, HS, BS, BI;
	int RW, RH;
	Glib::Property<Glib::ustring> propertyTitle;
	Glib::Property<Glib::ustring> propertyIconName;
	PGdkPixbuf icon;

};

class DesignDialog : public DesignWindow {

public:
	DesignDialog();
	~DesignDialog();

	Glib::PropertyProxy<bool> property_has_separator() { return propertyHasSeparator.get_proxy(); }

	Gtk::Bin * getBin() { return &bin; }
	Gtk::HButtonBox * getButtons() { return &buttons; }

protected:
	void onHasSeparatorChanged();
	void onPropertyChanged(const char * name);

private:
	Gtk::VBox box;
	Gtk::Alignment bin;
	Gtk::HSeparator separator;
	Gtk::Alignment sepalign;
	Gtk::HButtonBox buttons;
	Glib::Property<bool> propertyHasSeparator;

};

class DesignColorSelectionDialog : public DesignDialog {

public:
	DesignColorSelectionDialog();

	Gtk::ColorSelection * getColorsel() { return &colorsel; }

private:
	Gtk::ColorSelection colorsel;
	Gtk::Button okButton, cancelButton;

};

class DesignFileChooserDialog : public DesignDialog {

public:
	DesignFileChooserDialog();

	Glib::PropertyProxy<Gtk::FileChooserAction> property_action() { return propertyAction.get_proxy(); }
	Glib::PropertyProxy<Gtk::Widget *> property_extra_widget() { return propertyExtraWidget.get_proxy(); }
	Glib::PropertyProxy<bool> property_local_only() { return propertyLocalOnly.get_proxy(); }
	Glib::PropertyProxy<Gtk::Widget *> property_preview_widget() { return propertyPreviewWidget.get_proxy(); }
	Glib::PropertyProxy<bool> property_preview_widget_active() { return propertyPreviewWidgetActive.get_proxy(); }
	Glib::PropertyProxy<bool> property_select_multiple() { return propertySelectMultiple.get_proxy(); }
	Glib::PropertyProxy<bool> property_show_hidden() { return propertyShowHidden.get_proxy(); }
	Glib::PropertyProxy<bool> property_use_preview_label() { return propertyUsePreviewLabel.get_proxy(); }
	Glib::PropertyProxy<bool> property_do_overwrite_confirmation() { return propertyDoOverwriteConfirmation.get_proxy(); }

	Gtk::FileChooserWidget * getFileChooser() { return &fileChooser; }

private:
	Glib::Property<Gtk::FileChooserAction> propertyAction;
	Glib::Property<Gtk::Widget *> propertyExtraWidget;
	Glib::Property<bool> propertyLocalOnly;
	Glib::Property<Gtk::Widget *> propertyPreviewWidget;
	Glib::Property<bool> propertyPreviewWidgetActive;
	Glib::Property<bool> propertySelectMultiple;
	Glib::Property<bool> propertyShowHidden;
	Glib::Property<bool> propertyUsePreviewLabel;
	Glib::Property<bool> propertyDoOverwriteConfirmation;

	Gtk::FileChooserWidget fileChooser;

};

class DesignRecentChooserDialog : public DesignDialog {

public:
	DesignRecentChooserDialog();

	Glib::PropertyProxy<int> property_limit() { return propertyLimit.get_proxy(); }
	Glib::PropertyProxy<bool> property_local_only() { return propertyLocalOnly.get_proxy(); }
	Glib::PropertyProxy<bool> property_select_multiple() { return propertySelectMultiple.get_proxy(); }
	Glib::PropertyProxy<bool> property_show_icons() { return propertyShowIcons.get_proxy(); }
	Glib::PropertyProxy<bool> property_show_not_found() { return propertyShowNotFound.get_proxy(); }
	Glib::PropertyProxy<bool> property_show_private() { return propertyShowPrivate.get_proxy(); }
	Glib::PropertyProxy<bool> property_show_tips() { return propertyShowTips.get_proxy(); }
	Glib::PropertyProxy<Gtk::RecentSortType> property_sort_type() { return propertySortType.get_proxy(); }

	Gtk::RecentChooserWidget * getRecentChooser() { return &recentChooser; }

private:
	Glib::Property<int> propertyLimit;
	Glib::Property<bool> propertyLocalOnly;
	Glib::Property<bool> propertySelectMultiple;
	Glib::Property<bool> propertyShowIcons;
	Glib::Property<bool> propertyShowNotFound;
	Glib::Property<bool> propertyShowPrivate;
	Glib::Property<bool> propertyShowTips;
	Glib::Property<Gtk::RecentSortType> propertySortType;

	Gtk::RecentChooserWidget recentChooser;

};

class DesignFontSelectionDialog : public DesignDialog {

public:
	DesignFontSelectionDialog();

	Gtk::FontSelection * getFontSelection() { return &fontSelection; }

private:
	Gtk::FontSelection fontSelection;
	Gtk::Button okButton, cancelButton;

};

class Button : public Gtk::Button {

public:
	Button();
	~Button();

};

class Custom : public Gtk::DrawingArea {

public:
	Custom();

protected:
	bool on_expose_event(GdkEventExpose * event);
	void on_size_request(Gtk::Requisition * requisition);

};

class ArrowButton : public Gtk::Button {

public:
	ArrowButton();

private:
	Gtk::Arrow arrow;

};

class EditorWidget : public virtual Gtk::Alignment {

public:
	EditorWidget();
	PEditor getEditor();
	void setEditor(PEditor editor);
	void unsetEditor();

	bool isFrameNeeded() { return frameNeeded; }

protected:
	void setFrameNeeded(bool fn) { frameNeeded= fn; }

private:
	PEditor editor;
	bool frameNeeded;

};

class HierarchyEditorWidget : public EditorWidget {

public:
	HierarchyEditorWidget();

	Polytree * getTreeView() { return &treeView; }

	int getSplit();
	void setSplit(int split);

	void collapse();

private:
	Gtk::ScrolledWindow scrolledWindow;
	Polytree treeView;

};

class PropertyEditorWidget : public EditorWidget, public virtual Polycell {

public:
	PropertyEditorWidget();
	bool isEditable();

};

class PropertyEditorWidgetText : public PropertyEditorWidget, public PolycellText {

public:
	PropertyEditorWidgetText();

};

class PropertyEditorWidgetInput : public PropertyEditorWidget, public PolycellInput {

public:
	PropertyEditorWidgetInput();

};

class PropertyEditorWidgetChoice : public PropertyEditorWidget, public PolycellChoice {

public:
	PropertyEditorWidgetChoice();

};

class ColorPropertyEditorWidget : public PropertyEditorWidget, public PolycellWidget {

public:
	ColorPropertyEditorWidget();

	const Gdk::Color * getColor() { return &color; }
	void setColor(const Gdk::Color & color);

protected:
	bool onDrawingExpose(GdkEventExpose * event);
	void onButtonClicked();

	void onEditingStarted();

	void cellRender(
		Gtk::Widget * widget,
		PGdkWindow window,
		const Gdk::Rectangle & rect,
		Gtk::CellRendererState flags
	);

	void paint(PGdkWindow window, const Gdk::Rectangle & rect);

private:
	Gtk::HBox hBox;
	Gtk::DrawingArea drawing;
	ArrowButton button;
	Gdk::Color color;

};

}

#endif
