/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2009 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#include "headers.h"
#include "util.h"

namespace CrowApp {

void CheckFailed(const char * expr, const char * file, int line) {
	std::ostringstream oss;
	oss << boost::format("Check failed: %s, file: %s, line: %d") % expr % file % line;
	FatalError(oss.str());
}

void FatalError(const std::string & message) {
	g_critical(message.c_str());
	#ifdef CROW_SANDBOX
		assert(0);
	#endif
	Gtk::MessageDialog dialog(
		"Fatal Error [" G_LOG_DOMAIN "]", false,
		Gtk::MESSAGE_ERROR, Gtk::BUTTONS_CLOSE, true
	);
	dialog.set_secondary_text(message);
	dialog.set_keep_above(true);
	dialog.run();
	abort();
}

std::vector<std::string> Arguments(int argc, char ** argv) {
	std::vector<std::string> arguments;
	arguments.reserve(argc);
	for(int i=0; i<argc; ++i)
		arguments.push_back(argv[i]);
	return arguments;
}

StringList Split(const std::string & str, const char * dlm) {
	StringList result;
	typedef boost::tokenizer<boost::char_separator<char> > Tok;
	boost::char_separator<char> sep(dlm, "", boost::keep_empty_tokens);
	Tok tok(str, sep);
	for(Tok::iterator it= tok.begin(); it!=tok.end(); ++it)
		result.push_back(*it);
	return result;
}

GHashTable * SettingsToHashTable(const Settings & settings) {
	GHashTable * table= g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
	for(Settings::const_iterator it= settings.begin(); it!=settings.end(); ++it) {
		char * key= g_strdup(it->first.c_str());
		char * value= g_strdup(it->second.c_str());
		g_hash_table_insert(table, key, value);
	}
	return table;
}

Settings HashTableToSettings(GHashTable * table) {
	Settings result;
	struct tmp { static void add(gpointer key, gpointer value, gpointer user_data) {
		(*static_cast<Settings *>(user_data))[static_cast<char *>(key)]= static_cast<char *>(value);
	} };
	g_hash_table_foreach(table, &tmp::add, &result);
	return result;
}

int GetKeyInt(const Settings & settings, const std::string & key) {
	Settings::const_iterator it= settings.find(key);
	CHECK(it!=settings.end());
	std::istringstream iss(it->second);
	int value;
	iss >> value;
	CHECK(!iss.fail());
	return value;
}

void SetKeyInt(Settings * settings, const std::string & key, int value) {
	std::ostringstream oss;
	oss << value;
	(*settings)[key]= oss.str();
}

Glib::ustring GetKeyString(const Settings & settings, const std::string & key) {
	Settings::const_iterator it= settings.find(key);
	CHECK(it!=settings.end());
	return it->second;
}

void SetKeyString(Settings * settings, const std::string & key, const char * const value) {
	(*settings)[key]= value;
}

void EnsureKeyString(Settings * settings, const std::string & key, const char * const value) {
	Settings::const_iterator it= settings->find(key);
	if(it==settings->end())
		(*settings)[key]= value;
}

bool GetKeyBool(const Settings & settings, const std::string & key) {
	Settings::const_iterator it= settings.find(key);
	CHECK(it!=settings.end());
	return it->second == "true" ? true : false;
}

void SetKeyBool(Settings * settings, const std::string & key, bool value) {
	(*settings)[key]= value ? "true" : "false";
}

Gdk::Color GetKeyColor(const Settings & settings, const std::string & key) {
	Settings::const_iterator it= settings.find(key);
	CHECK(it!=settings.end());
	return Gdk::Color(it->second);
}

void SetKeyColor(Settings * settings, const std::string & key, const Gdk::Color & value) {
	guint32 pixel=
		((value.get_red() >> 8) << 16) |
		((value.get_green() >> 8) << 8) |
		(value.get_blue() >> 8);
	std::ostringstream oss;
	oss << "#";
	oss << std::setw(6) << std::setfill('0');
	oss << std::hex << std::uppercase;
	oss << pixel;
	(*settings)[key]= oss.str();
}

}
